
# Imports
import os
import time
from typing import Any

from ..config import StouputilsConfig as Cfg

# Backwards compatibility for constants
RESET: str      = Cfg.RESET
RED: str        = Cfg.RED
GREEN: str      = Cfg.GREEN
YELLOW: str     = Cfg.YELLOW
BLUE: str       = Cfg.BLUE
MAGENTA: str    = Cfg.MAGENTA
CYAN: str       = Cfg.CYAN
LINE_UP: str    = Cfg.LINE_UP
BOLD: str       = Cfg.BOLD
BAR_FORMAT: str = Cfg.BAR_FORMAT

# Enable colors on Windows 10 terminal if applicable
if os.name == "nt":
	import subprocess
	subprocess.run("color", shell=True)

# Shared memory for print functions
class PrintMemory:
	""" Class to store shared variables for print functions. """
	previous_args_kwards: tuple[Any, Any] = ((), {})
	""" Store the previous print call's args and kwargs for comparison in is_same_print() """
	nb_values: int = 1
	""" Count how many times the same value has been printed in a row for is_same_print() """
	import_time: float = time.time()
	""" Store the import time to determine how to format timestamps in current_time() """





# Imports
import os
from typing import IO, Any


# Functions
def safe_close(file: IO[Any] | int | Any | None) -> None:
	""" Safely close a file object (or file descriptor) after flushing, ignoring any exceptions.

	Args:
		file (IO[Any] | int | None): The file object or file descriptor to close
	"""
	if isinstance(file, int):
		if file != -1:
			for func in (os.fsync, os.close):
				try:
					func(file)
				except Exception:
					pass
	elif file:
		for func in ("flush", "close"):
			try:
				getattr(file, func)()
			except Exception:
				pass




""" Installer module for Linux/macOS specific functions.

Provides Linux/macOS specific implementations for checking admin privileges,
determining appropriate installation paths (global/local), and suggesting
how to add directories to the system's PATH environment variable.
"""
# Imports
import os

from ..decorators import LogLevels, handle_error
from ..io.path import clean_path
from ..print.message import debug, info, warning
from .common import ask_install_type, prompt_for_path


# Functions
@handle_error(message="Failed to suggest how to add to PATH (Linux)", error_log=LogLevels.WARNING_TRACEBACK)
def add_to_path_linux(install_path: str) -> bool:
	""" Suggest how to add install_path to PATH on Linux.

	Checks the current shell and provides instructions for adding the path
	to the appropriate configuration file (e.g., .bashrc, .zshrc, config.fish).

	Args:
		install_path (str): The path to add to the PATH environment variable.

	Returns:
		bool: True if instructions were provided, False otherwise (e.g., unknown shell).
	"""
	shell_config_files: dict[str, str] = {
		"bash": "~/.bashrc",
		"zsh": "~/.zshrc",
		"fish": "~/.config/fish/config.fish"
	}
	current_shell: str = os.environ.get("SHELL", "").split('/')[-1]
	config_file: str | None = shell_config_files.get(current_shell)

	if config_file:
		export_cmd: str = ""
		if current_shell == "fish":
			export_cmd = f"set -gx PATH $PATH {install_path}"
		else:
			export_cmd = f"export PATH=\"$PATH:{install_path}\"" # Escape quotes for print

		debug(
			f"To add the installation directory to your PATH, add the following line to your '{config_file}':\n"
			f"  {export_cmd}\n"
			f"Then restart your shell or run 'source {config_file}'."
		)
		return True
	else:
		warning(f"Could not determine your shell configuration file. Please add '{install_path}' to your PATH manually.")
		return False


def check_admin_linux() -> bool:
	""" Check if the script is running with root privileges on Linux/macOS.

	Returns:
		bool: True if the effective user ID is 0 (root), False otherwise.
	"""
	try:
		return os.geteuid() == 0 # type: ignore
	except AttributeError as e:
		# os.geteuid() is not available on all platforms (e.g., Windows)
		# This function should ideally only be called on Linux/macOS.
		warning(f"Could not determine user privileges on this platform: {e}")
		return False
	except Exception as e:
		warning(f"Error checking admin privileges: {e}")
		return False


@handle_error(message="Failed to get installation path (Linux)", error_log=LogLevels.ERROR_TRACEBACK)
def get_install_path_linux(
	program_name: str,
	ask_global: int = 0,
	add_path: bool = True,
	append_to_path: str = "",
	default_global: str = "/usr/local/bin",
	) -> str:
	""" Get the installation path for the program on Linux/macOS.

	Args:
		program_name   (str):   The name of the program to install.
		ask_global     (int):   0 = ask for anything, 1 = install globally, 2 = install locally
		add_path       (bool):  Whether to add the program to the PATH environment variable. (Only if installed globally)
		append_to_path (str):   String to append to the installation path when adding to PATH.
			(ex: "bin" if executables are in the bin folder)
		default_global (str):   The default global installation path.
			(Default is "/usr/local/bin" which is the most common location for executables on Linux/macOS,
			could be "/opt" or any other directory)

	Returns:
		str: The chosen installation path, or an empty string if installation is cancelled.
	"""
	# Default paths
	default_local_path: str = clean_path(os.path.join(os.getcwd(), program_name))

	# Common global locations: /usr/local/bin for executables, /opt/ for self-contained apps
	# We assume 'program_name' might be an executable or a directory, /usr/local/ is safer
	default_global_path: str = clean_path(f"{default_global}/{program_name}") # Or potentially /opt/{program_name}

	# Ask user for installation type (global/local)
	install_type: str = ask_install_type(ask_global, default_local_path, default_global_path)

	# Handle global installation choice
	if install_type == 'g':
		if not check_admin_linux():
			warning(
				f"Global installation typically requires sudo privileges to write to "
				f"'{os.path.dirname(default_global_path)}'.\n"
				f"You may need to re-run the script with 'sudo'.\n"
				f"Install locally instead to '{default_local_path}'? (Y/n): "
			)
			if input().lower() == 'n':
				info("Installation cancelled.")
				return ""
			else:
				# Fallback to local path if user agrees
				return prompt_for_path(
					f"Falling back to local installation path: {default_local_path}.",
					default_local_path
				)
		else:
			# User is admin or proceeding with global install anyway
			install_path: str = prompt_for_path(
				f"Default global installation path is {default_global_path}.",
				default_global_path
			)
			if add_path:
				# Suggest adding the *directory* containing the program to PATH,
				# or the path itself if it seems like a directory install
				path_to_add: str = os.path.dirname(install_path) if os.path.isfile(install_path) else install_path
				add_to_path_linux(os.path.join(path_to_add, append_to_path))
			return install_path

	# Handle local installation choice
	else: # install_type == 'l'
		return prompt_for_path(
			f"Default local installation path is {default_local_path}.",
			default_local_path
		)





# Imports
import os
import shutil
import zipfile

from ..config import StouputilsConfig as Cfg
from ..decorators import measure_time
from ..io.path import clean_path
from ..print.message import info, warning
from ..print.progress_bar import colored_for_loop
from .retrieve import get_all_previous_backups


# Function to consolidate multiple backups into one comprehensive backup
@measure_time(message="Consolidating backups")
def consolidate_backups(zip_path: str, destination_zip: str) -> None:
	""" Consolidates the files from the given backup and all previous ones into a new ZIP file,
	ensuring that the most recent version of each file is kept and deleted files are not restored.

	Args:
		zip_path (str): Path to the latest backup ZIP file (If endswith "/latest.zip" or "/", the latest backup will be used)
		destination_zip (str): Path to the destination ZIP file where the consolidated backup will be saved
	Examples:

	.. code-block:: python

		> consolidate_backups("/path/to/backups/latest.zip", "/path/to/consolidated.zip")
		[INFO HH:MM:SS] Consolidating backups
		[INFO HH:MM:SS] Consolidated backup created: '/path/to/consolidated.zip'
	"""
	zip_path = clean_path(os.path.abspath(zip_path))
	destination_zip = clean_path(os.path.abspath(destination_zip))
	zip_folder: str = clean_path(os.path.dirname(zip_path))

	# Get all previous backups up to the specified one
	previous_backups: dict[str, dict[str, str]] = get_all_previous_backups(zip_folder, all_before=zip_path)
	backup_paths: list[str] = list(previous_backups.keys())

	# First pass: collect all deleted files and build file registry
	deleted_files: set[str] = set()
	file_registry: dict[str, tuple[str, zipfile.ZipInfo]] = {}  # filename -> (backup_path, zipinfo)

	# Process backups in reverse order (newest first) to prioritize latest versions
	for backup_path in reversed(backup_paths):
		try:
			with zipfile.ZipFile(backup_path, "r") as zipf_in:

				# Get namelist once for efficiency
				namelist: list[str] = zipf_in.namelist()

				# Process deleted files
				if "__deleted_files__.txt" in namelist:
					backup_deleted_files: list[str] = zipf_in.read("__deleted_files__.txt").decode().splitlines()
					deleted_files.update(backup_deleted_files)

				# Process files - only add if not already in registry (newer versions take precedence)
				for inf in zipf_in.infolist():
					filename: str = inf.filename
					if (filename
						and filename != "__deleted_files__.txt"
						and filename not in deleted_files
						and filename not in file_registry):
						file_registry[filename] = (backup_path, inf)
		except Exception as e:
			warning(f"Error processing backup {backup_path}: {e}")
			continue

	# Second pass: copy files efficiently, keeping ZIP files open longer
	open_zips: dict[str, zipfile.ZipFile] = {}

	try:
		with zipfile.ZipFile(destination_zip, "w", compression=zipfile.ZIP_DEFLATED, compresslevel=9) as zipf_out:
			for filename, (backup_path, inf) in colored_for_loop(file_registry.items(), desc="Making consolidated backup"):
				try:
					# Open ZIP file if not already open
					if backup_path not in open_zips:
						open_zips[backup_path] = zipfile.ZipFile(backup_path, "r")

					zipf_in = open_zips[backup_path]

					# Copy file with optimized strategy based on file size
					with zipf_in.open(inf, "r") as source:
						with zipf_out.open(inf, "w", force_zip64=True) as target:
							# Use shutil.copyfileobj with larger chunks for files >50MB
							if inf.file_size > 52428800:  # 50MB threshold
								shutil.copyfileobj(source, target, length=Cfg.LARGE_CHUNK_SIZE)
							else:
								# Use shutil.copyfileobj with standard chunks for smaller files
								shutil.copyfileobj(source, target, length=Cfg.CHUNK_SIZE)
				except Exception as e:
					warning(f"Error copying file {filename} from {backup_path}: {e}")
					continue

			# Add accumulated deleted files to the consolidated backup
			if deleted_files:
				zipf_out.writestr("__deleted_files__.txt", "\n".join(sorted(deleted_files)), compress_type=zipfile.ZIP_DEFLATED)
	finally:
		# Clean up open ZIP files
		for zipf in open_zips.values():
			try:
				zipf.close()
			except Exception:
				pass

	info(f"Consolidated backup created: {destination_zip}")




"""
This module provides decorators for various purposes:

- :py:deco:`measure_time` - Measure the execution time of a function and print it with the given print function
- :py:deco:`handle_error` - Handle an error with different log levels
- :py:deco:`timeout` - Raise an exception if the function runs longer than the specified timeout
- :py:deco:`retry` - Retry a function when specific exceptions are raised, with configurable delay and max attempts
- :py:deco:`simple_cache` - Easy cache function with parameter caching method
- :py:deco:`abstract` - Mark a function as abstract, using :py:class:`~handle_error.LogLevels` for error handling
- :py:deco:`deprecated` - Mark a function as deprecated, using :py:class:`~handle_error.LogLevels` for warning handling
- :py:deco:`silent` - Make a function silent (disable stdout, and stderr if specified) (alternative to :py:class:`stouputils.ctx.Muffle`)

.. image:: https://raw.githubusercontent.com/Stoupy51/stouputils/refs/heads/main/assets/decorators_module_1.gif
  :alt: stouputils decorators examples

.. image:: https://raw.githubusercontent.com/Stoupy51/stouputils/refs/heads/main/assets/decorators_module_2.gif
  :alt: stouputils decorators examples
"""

# Imports
from .abstract import *
from .common import *
from .deprecated import *
from .handle_error import *
from .measure_time import *
from .retry import *
from .silent import *
from .simple_cache import *
from .timeout import *




from ..ctx.set_mp_start_method import SetMPStartMethod as SetMPStartMethod
from ..typing import JsonList as JsonList
from .capturer import CaptureOutput as CaptureOutput
from .common import handle_parameters as handle_parameters, nice_wrapper as nice_wrapper, resolve_process_title as resolve_process_title
from collections.abc import Callable as Callable, Iterable
from typing import Any

def doctest_square(x: int) -> int: ...
def doctest_slow(x: int) -> int: ...
def multiprocessing[T, R](func: Callable[..., R] | list[Callable[..., R]], args: Iterable[T], use_starmap: bool = False, chunksize: int = 1, desc: str = '', max_workers: int | float = ..., capture_output: bool = False, delay_first_calls: float = 0, nice: int | None = None, process_title: str | None = None, color: str = ..., bar_format: str = ..., ascii: bool = False, smooth_tqdm: bool = True, **tqdm_kwargs: Any) -> list[R]:
    ''' Method to execute a function in parallel using multiprocessing

\t- For CPU-bound operations where the GIL (Global Interpreter Lock) is a bottleneck.
\t- When the task can be divided into smaller, independent sub-tasks that can be executed concurrently.
\t- For computationally intensive tasks like scientific simulations, data analysis, or machine learning workloads.

\tArgs:
\t\tfunc\t\t\t\t(Callable | list[Callable]):\tFunction to execute, or list of functions (one per argument)
\t\targs\t\t\t\t(Iterable):\t\t\tIterable of arguments to pass to the function(s)
\t\tuse_starmap\t\t\t(bool):\t\t\t\tWhether to use starmap or not (Defaults to False):
\t\t\tTrue means the function will be called like func(\\*args[i]) instead of func(args[i])
\t\tchunksize\t\t\t(int):\t\t\t\tNumber of arguments to process at a time
\t\t\t(Defaults to 1 for proper progress bar display)
\t\tdesc\t\t\t\t(str):\t\t\t\tDescription displayed in the progress bar
\t\t\t(if not provided no progress bar will be displayed)
\t\tmax_workers\t\t\t(int | float):\t\tNumber of workers to use (Defaults to CPU_COUNT), -1 means CPU_COUNT.
\t\t\tIf float between 0 and 1, it\'s treated as a percentage of CPU_COUNT.
\t\t\tIf negative float between -1 and 0, it\'s treated as a percentage of len(args).
\t\tcapture_output\t\t(bool):\t\t\t\tWhether to capture stdout/stderr from the worker processes (Defaults to True)
\t\tdelay_first_calls\t(float):\t\t\tApply i*delay_first_calls seconds delay to the first "max_workers" calls.
\t\t\tFor instance, the first process will be delayed by 0 seconds, the second by 1 second, etc.
\t\t\t(Defaults to 0): This can be useful to avoid functions being called in the same second.
\t\tnice\t\t\t\t(int | None):\t\tAdjust the priority of worker processes (Defaults to None).
\t\t\tUse Unix-style values: -20 (highest priority) to 19 (lowest priority).
\t\t\tPositive values reduce priority, negative values increase it.
\t\t\tAutomatically converted to appropriate priority class on Windows.
\t\t\tIf None, no priority adjustment is made.
\t\tprocess_title\t\t(str | None):\t\tIf provided, sets the process title for worker processes.
\t\t\tIf it starts with \'+++\', this prefix is replaced by the current process title.
\t\tcolor\t\t\t\t(str):\t\t\t\tColor of the progress bar (Defaults to MAGENTA)
\t\tbar_format\t\t\t(str):\t\t\t\tFormat of the progress bar (Defaults to BAR_FORMAT)
\t\tascii\t\t\t\t(bool):\t\t\t\tWhether to use ASCII or Unicode characters for the progress bar
\t\tsmooth_tqdm\t\t\t(bool):\t\t\t\tWhether to enable smooth progress bar updates by setting miniters and mininterval (Defaults to True)
\t\t**tqdm_kwargs\t\t(Any):\t\t\t\tAdditional keyword arguments to pass to tqdm

\tReturns:
\t\tlist[object]:\tResults of the function execution

\tExamples:
\t\t.. code-block:: python

\t\t\t> multiprocessing(doctest_square, args=[1, 2, 3])
\t\t\t[1, 4, 9]

\t\t\t> multiprocessing(int.__mul__, [(1,2), (3,4), (5,6)], use_starmap=True)
\t\t\t[2, 12, 30]

\t\t\t> # Using a list of functions (one per argument)
\t\t\t> multiprocessing([doctest_square, doctest_square, doctest_square], [1, 2, 3])
\t\t\t[1, 4, 9]

\t\t\t> # Will process in parallel with progress bar
\t\t\t> multiprocessing(doctest_slow, range(10), desc="Processing")
\t\t\t[0, 1, 2, 3, 4, 5, 6, 7, 8, 9]

\t\t\t> # Will process in parallel with progress bar and delay the first threads
\t\t\t> multiprocessing(
\t\t\t.     doctest_slow,
\t\t\t.     range(10),
\t\t\t.     desc="Processing with delay",
\t\t\t.     max_workers=2,
\t\t\t.     delay_first_calls=0.6,
\t\t\t.     process_title="+++ (Worker)"
\t\t\t. )
\t\t\t[0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
\t'''
def multithreading[T, R](func: Callable[..., R] | list[Callable[..., R]], args: Iterable[T], use_starmap: bool = False, desc: str = '', max_workers: int | float = ..., delay_first_calls: float = 0, color: str = ..., bar_format: str = ..., ascii: bool = False, smooth_tqdm: bool = True, **tqdm_kwargs: Any) -> list[R]:
    ''' Method to execute a function in parallel using multithreading, you should use it:

\t- For I/O-bound operations where the GIL is not a bottleneck, such as network requests or disk operations.
\t- When the task involves waiting for external resources, such as network responses or user input.
\t- For operations that involve a lot of waiting, such as GUI event handling or handling user input.

\tArgs:
\t\tfunc\t\t\t\t(Callable | list[Callable]):\tFunction to execute, or list of functions (one per argument)
\t\targs\t\t\t\t(Iterable):\t\t\tIterable of arguments to pass to the function(s)
\t\tuse_starmap\t\t\t(bool):\t\t\t\tWhether to use starmap or not (Defaults to False):
\t\t\tTrue means the function will be called like func(\\*args[i]) instead of func(args[i])
\t\tdesc\t\t\t\t(str):\t\t\t\tDescription displayed in the progress bar
\t\t\t(if not provided no progress bar will be displayed)
\t\tmax_workers\t\t\t(int | float):\t\tNumber of workers to use (Defaults to CPU_COUNT), -1 means CPU_COUNT.
\t\t\tIf float between 0 and 1, it\'s treated as a percentage of CPU_COUNT.
\t\t\tIf negative float between -1 and 0, it\'s treated as a percentage of len(args).
\t\tdelay_first_calls\t(float):\t\t\tApply i*delay_first_calls seconds delay to the first "max_workers" calls.
\t\t\tFor instance with value to 1, the first thread will be delayed by 0 seconds, the second by 1 second, etc.
\t\t\t(Defaults to 0): This can be useful to avoid functions being called in the same second.
\t\tcolor\t\t\t\t(str):\t\t\t\tColor of the progress bar (Defaults to MAGENTA)
\t\tbar_format\t\t\t(str):\t\t\t\tFormat of the progress bar (Defaults to BAR_FORMAT)
\t\tascii\t\t\t\t(bool):\t\t\t\tWhether to use ASCII or Unicode characters for the progress bar
\t\tsmooth_tqdm\t\t\t(bool):\t\t\t\tWhether to enable smooth progress bar updates by setting miniters and mininterval (Defaults to True)
\t\t**tqdm_kwargs\t\t(Any):\t\t\t\tAdditional keyword arguments to pass to tqdm

\tReturns:
\t\tlist[object]:\tResults of the function execution

\tExamples:
\t\t.. code-block:: python

\t\t\t> multithreading(doctest_square, args=[1, 2, 3])
\t\t\t[1, 4, 9]

\t\t\t> multithreading(int.__mul__, [(1,2), (3,4), (5,6)], use_starmap=True)
\t\t\t[2, 12, 30]

\t\t\t> # Using a list of functions (one per argument)
\t\t\t> multithreading([doctest_square, doctest_square, doctest_square], [1, 2, 3])
\t\t\t[1, 4, 9]

\t\t\t> # Will process in parallel with progress bar
\t\t\t> multithreading(doctest_slow, range(10), desc="Threading")
\t\t\t[0, 1, 2, 3, 4, 5, 6, 7, 8, 9]

\t\t\t> # Will process in parallel with progress bar and delay the first threads
\t\t\t> multithreading(
\t\t\t.     doctest_slow,
\t\t\t.     range(10),
\t\t\t.     desc="Threading with delay",
\t\t\t.     max_workers=2,
\t\t\t.     delay_first_calls=0.6
\t\t\t. )
\t\t\t[0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
\t'''
def capture_subprocess_output[T, R](args: tuple[CaptureOutput, Callable[[T], R], T]) -> R:
    """ Wrapper function to execute the target function in a subprocess with optional output capture.

\tArgs:
\t\ttuple[CaptureOutput,Callable,T]: Tuple containing:
\t\t\tCaptureOutput: Capturer object to redirect stdout/stderr
\t\t\tCallable: Target function to execute
\t\t\tT: Argument to pass to the target function
\t"""
def process_title_wrapper[T, R](args: tuple[str, int, Callable[[T], R], T]) -> R:
    """ Wrapper function to set the process title before executing the target function.

\tBehavior depends on StouputilsConfig.PROCESS_TITLE_PER_WORKER:
\t- If True: Title is set only once per worker, index reflects worker number (0 to max_workers-1)
\t- If False: Title is updated for each task, index reflects task number (0 to len(args)-1)

\tArgs:
\t\ttuple[str,int,Callable,T]: Tuple containing:
\t\t\tstr: Process title to set
\t\t\tint: Worker index to append to title
\t\t\tCallable: Target function to execute
\t\t\tT: Argument to pass to the target function
\t"""



import numpy as np
from ..io.path import super_open as super_open
from ..print.message import debug as debug, info as info
from numpy.typing import NDArray as NDArray

def numpy_to_obj(path: str, array: NDArray[np.integer | np.floating | np.bool_], threshold: float = 0.5, step_size: int = 1, pad_array: bool = True, verbose: int = 0) -> None:
    ''' Generate a \'.obj\' file from a numpy array for 3D visualization using marching cubes.

\tArgs:
\t\tpath      (str):     Path to the output .obj file.
\t\tarray     (NDArray): Numpy array to be dumped (must be 3D).
\t\tthreshold (float):   Threshold level for marching cubes (0.5 for binary data).
\t\tstep_size (int):     Step size for marching cubes (higher = simpler mesh, faster generation).
\t\tpad_array (bool):    If True, pad array with zeros to ensure closed volumes for border cells.
\t\tverbose   (int):     Verbosity level (0 = no output, 1 = some output, 2 = full output).

\tExamples:

\t\t.. code-block:: python

\t\t\t> array = np.random.rand(64, 64, 64) > 0.5  # Binary volume
\t\t\t> numpy_to_obj("output_mesh.obj", array, threshold=0.5, step_size=2, pad_array=True, verbose=1)

\t\t\t> array = my_3d_data  # Some 3D numpy array (e.g. human lung scan)
\t\t\t> numpy_to_obj("output_mesh.obj", array, threshold=0.3)
\t'''





# PYTHON_ARGCOMPLETE_OK
# Imports
import argparse
import sys

import argcomplete

from .decorators.handle_error import handle_error

# Argument Parser Setup for Auto-Completion
parser = argparse.ArgumentParser(prog="stouputils", add_help=False)
parser.add_argument("command", nargs="?", choices=[
	"--version", "-v", "version", "show_version", "all_doctests", "archive", "backup", "build", "changelog", "redirect"
])
parser.add_argument("args", nargs="*")
argcomplete.autocomplete(parser)


@handle_error(message="Error while running 'stouputils'")
def main() -> None:
	second_arg: str = sys.argv[1].lower() if len(sys.argv) >= 2 else ""

	# Print the version of stouputils and its dependencies
	if second_arg in ("--version", "-v", "version", "show_version"):
		from .version_pkg import show_version_cli
		return show_version_cli()

	# Handle "all_doctests" command
	if second_arg.replace("-", "_").startswith("all_doctest"):
		root_dir: str = "." if len(sys.argv) == 2 else sys.argv[2]
		pattern: str = sys.argv[3] if len(sys.argv) >= 4 else "*"
		from .all_doctests.main import launch_tests
		if launch_tests(root_dir, pattern=pattern) > 0:
			sys.exit(1)
		return

	# Handle "archive" command
	if second_arg == "archive":
		sys.argv.pop(1)  # Remove "archive" from argv so archive_cli gets clean arguments
		from .archive import archive_cli
		return archive_cli()

	# Handle "backup" command
	if second_arg == "backup":
		sys.argv.pop(1)  # Remove "backup" from argv so backup_cli gets clean arguments
		from .backup.cli import backup_cli
		return backup_cli()

	# Handle "build" command
	if second_arg == "build":
		from .continuous_delivery.pypi import pypi_full_routine_using_uv
		return pypi_full_routine_using_uv()

	# Handle "changelog" command
	if second_arg == "changelog":
		sys.argv.pop(1)  # Remove "changelog" from argv so changelog_cli gets clean arguments
		from .continuous_delivery.git import changelog_cli
		return changelog_cli()

	# Handle "redirect" command
	if second_arg == "redirect":
		sys.argv.pop(1)  # Remove "redirect" from argv so redirect_cli gets clean arguments
		from .io.redirect import redirect_cli
		return redirect_cli()

	# Get version
	from importlib.metadata import version
	try:
		pkg_version = version("stouputils")
	except Exception:
		pkg_version = "unknown"

	# Print help with nice formatting
	from .config import StouputilsConfig as Cfg
	separator: str = "─" * 60
	print(f"""
{Cfg.CYAN}{separator}{Cfg.RESET}
{Cfg.CYAN}stouputils {Cfg.GREEN}CLI {Cfg.CYAN}v{pkg_version}{Cfg.RESET}
{Cfg.CYAN}{separator}{Cfg.RESET}
{Cfg.CYAN}Usage:{Cfg.RESET} stouputils <command> [options]

{Cfg.CYAN}Available commands:{Cfg.RESET}
  {Cfg.GREEN}--version, -v{Cfg.RESET} [pkg] [-t <depth>]    Show version information (optionally for a specific package)
  {Cfg.GREEN}all_doctests{Cfg.RESET} [dir] [pattern]        Run all doctests in the specified directory (optionally filter by pattern)
  {Cfg.GREEN}archive{Cfg.RESET} --help                      Archive utilities (make, repair)
  {Cfg.GREEN}backup{Cfg.RESET} --help                       Backup utilities (delta, consolidate, limit)
  {Cfg.GREEN}build{Cfg.RESET} [--no_stubs] [<minor|major>]  Build and publish package to PyPI using 'uv' tool (complete routine)
  {Cfg.GREEN}changelog{Cfg.RESET} [mode] [value] [options]  Generate changelog from local git history (see --help for details)
  {Cfg.GREEN}redirect{Cfg.RESET} <src> <dst> [--hardlink|--symlink]  Move a folder and create a link at the original path
{Cfg.CYAN}{separator}{Cfg.RESET}
""".strip())
	return

if __name__ == "__main__":
	main()





# Imports
import os
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
	import numpy as np
	from numpy.typing import NDArray

# Functions
def numpy_to_gif(
	path: str,
	array: "NDArray[np.integer | np.floating | np.bool_]",
	duration: int = 100,
	loop: int = 0,
	mkdir: bool = True,
	**kwargs: Any
) -> None:
	""" Generate a '.gif' file from a numpy array for 3D/4D visualization.

	Args:
		path     (str):     Path to the output .gif file.
		array    (NDArray): Numpy array to be dumped (must be 3D or 4D).
			3D: (depth, height, width) - e.g. (64, 1024, 1024)
			4D: (depth, height, width, channels) - e.g. (50, 64, 1024, 3)
		duration (int):     Duration between frames in milliseconds.
		loop     (int):     Number of loops (0 = infinite).
		mkdir    (bool):    Create the directory if it does not exist.
		**kwargs (Any):     Additional keyword arguments for PIL.Image.save().

	Examples:

		.. code-block:: python

			> # 3D array example
			> array = np.random.randint(0, 256, (10, 100, 100), dtype=np.uint8)
			> numpy_to_gif("output_10_frames_100x100.gif", array, duration=200, loop=0)

			> # 4D array example (batch of 3D images)
			> array_4d = np.random.randint(0, 256, (5, 10, 100, 3), dtype=np.uint8)
			> numpy_to_gif("output_50_frames_100x100.gif", array_4d, duration=200)

			> total_duration = 1000  # 1 second
			> numpy_to_gif("output_1s.gif", array, duration=total_duration // len(array))
	"""
	# Imports
	import numpy as np
	from PIL import Image

	# Assertions
	assert array.ndim in (3, 4), f"The input array must be 3D or 4D, got shape {array.shape} instead."
	if array.ndim == 4:
		assert array.shape[-1] in (1, 3), f"For 4D arrays, the last dimension must be 1 or 3 (channels), got shape {array.shape} instead."

	# Create directory if needed
	if mkdir:
		dirname: str = os.path.dirname(path)
		if dirname != "":
			os.makedirs(dirname, exist_ok=True)

	# Normalize array if outside [0-255] range to [0-1]
	array = array.astype(np.float32)
	mini, maxi = np.min(array), np.max(array)
	if mini < 0 or maxi > 255:
		array = ((array - mini) / (maxi - mini + 1e-8))

	# Scale to [0-255] if in [0-1] range
	mini, maxi = np.min(array), np.max(array)
	if mini >= 0.0 and maxi <= 1.0:
		array = (array * 255)

	# Ensure array is uint8 for PIL compatibility
	array = array.astype(np.uint8)

	# Convert each slice to PIL Image
	pil_images: list[Image.Image] = [
		Image.fromarray(z_slice)
		for z_slice in array
	]

	# Save as GIF
	pil_images[0].save(
		path,
		save_all=True,
		append_images=pil_images[1:],
		duration=duration,
		loop=loop,
		**kwargs
	)




import polars as pl
from typing import Any

def upsert_in_dataframe(df: pl.DataFrame, new_entry: dict[str, Any], primary_keys: list[str] | dict[str, Any] | None = None) -> pl.DataFrame:
    ''' Insert or update a row in the Polars DataFrame based on primary keys.

\tArgs:
\t\tdf\t\t\t\t(pl.DataFrame):\t\tThe Polars DataFrame to update.
\t\tnew_entry\t\t(dict[str, Any]):\tThe new entry to insert or update.
\t\tprimary_keys\t(list[str] | dict[str, Any] | None):\tThe primary keys to identify the row (for updates).
\tReturns:
\t\tpl.DataFrame: The updated Polars DataFrame.
\tExamples:
\t\t>>> import polars as pl
\t\t>>> df = pl.DataFrame({"id": [1, 2], "value": ["a", "b"]})
\t\t>>> new_entry = {"id": 2, "value": "updated"}
\t\t>>> updated_df = upsert_in_dataframe(df, new_entry, primary_keys=["id"])
\t\t>>> print(updated_df)
\t\tshape: (2, 2)
\t\t┌─────┬─────────┐
\t\t│ id  ┆ value   │
\t\t│ --- ┆ ---     │
\t\t│ i64 ┆ str     │
\t\t╞═════╪═════════╡
\t\t│ 1   ┆ a       │
\t\t│ 2   ┆ updated │
\t\t└─────┴─────────┘

\t\t>>> new_entry = {"id": 3, "value": "new"}
\t\t>>> updated_df = upsert_in_dataframe(updated_df, new_entry, primary_keys=["id"])
\t\t>>> print(updated_df)
\t\tshape: (3, 2)
\t\t┌─────┬─────────┐
\t\t│ id  ┆ value   │
\t\t│ --- ┆ ---     │
\t\t│ i64 ┆ str     │
\t\t╞═════╪═════════╡
\t\t│ 1   ┆ a       │
\t\t│ 2   ┆ updated │
\t\t│ 3   ┆ new     │
\t\t└─────┴─────────┘
\t'''




# Imports
from typing import Any


# Functions
def sort_dict_keys[T](dictionary: dict[T, Any], order: list[T], reverse: bool = False) -> dict[T, Any]:
	""" Sort dictionary keys using a given order list (reverse optional)

	Args:
		dictionary	(dict[T, Any]):	The dictionary to sort
		order		(list[T]):		The order list
		reverse		(bool):			Whether to sort in reverse order (given to sorted function which behaves differently than order.reverse())
	Returns:
		dict[T, Any]: The sorted dictionary

	Examples:
		>>> sort_dict_keys({'b': 2, 'a': 1, 'c': 3}, order=["a", "b", "c"])
		{'a': 1, 'b': 2, 'c': 3}

		>>> sort_dict_keys({'b': 2, 'a': 1, 'c': 3}, order=["a", "b", "c"], reverse=True)
		{'c': 3, 'b': 2, 'a': 1}

		>>> sort_dict_keys({'b': 2, 'a': 1, 'c': 3, 'd': 4}, order=["c", "b"])
		{'c': 3, 'b': 2, 'a': 1, 'd': 4}
	"""
	return dict(sorted(dictionary.items(), key=lambda x: order.index(x[0]) if x[0] in order else len(order), reverse=reverse))




from typing import Any, IO

def get_root_path(relative_path: str, go_up: int = 0) -> str:
    """ Get the absolute path of the directory.
\tUsually used to get the root path of the project using the __file__ variable.

\tArgs:
\t\trelative_path   (str): The path to get the absolute directory path from
\t\tgo_up           (int): Number of parent directories to go up (default: 0)
\tReturns:
\t\tstr: The absolute path of the directory

\tExamples:

\t\t.. code-block:: python

\t\t\t> get_root_path(__file__)
\t\t\t'C:/Users/Alexandre-PC/AppData/Local/Programs/Python/Python310/lib/site-packages/stouputils'

\t\t\t> get_root_path(__file__, 3)
\t\t\t'C:/Users/Alexandre-PC/AppData/Local/Programs/Python/Python310'
\t"""
def relative_path(file_path: str, relative_to: str = '') -> str:
    ''' Get the relative path of a file relative to a given directory.

\tArgs:
\t\tfile_path     (str): The path to get the relative path from
\t\trelative_to   (str): The path to get the relative path to (default: current working directory -> os.getcwd())
\tReturns:
\t\tstr: The relative path of the file
\tExamples:

\t\t>>> relative_path("D:/some/random/path/stouputils/io.py", "D:\\\\some")
\t\t\'random/path/stouputils/io.py\'
\t\t>>> relative_path("D:/some/random/path/stouputils/io.py", "D:\\\\some\\\\")
\t\t\'random/path/stouputils/io.py\'
\t'''
def super_copy(src: str, dst: str, create_dir: bool = True, symlink: bool = False) -> str:
    """ Copy a file (or a folder) from the source to the destination

\tArgs:
\t\tsrc         (str):  The source path
\t\tdst         (str):  The destination path
\t\tcreate_dir  (bool): Whether to create the directory if it doesn't exist (default: True)
\t\tsymlink     (bool): Whether to create a symlink instead of copying (Linux only)
\tReturns:
\t\tstr: The destination path
\t"""
def super_open(file_path: str, mode: str, encoding: str = 'utf-8') -> IO[Any]:
    ''' Open a file with the given mode, creating the directory if it doesn\'t exist (only if writing)

\tArgs:
\t\tfile_path\t(str): The path to the file
\t\tmode\t\t(str): The mode to open the file with, ex: "w", "r", "a", "wb", "rb", "ab"
\t\tencoding\t(str): The encoding to use when opening the file (default: "utf-8")
\tReturns:
\t\topen: The file object, ready to be used
\t'''
def read_file(file_path: str, encoding: str = 'utf-8') -> str:
    ''' Read the content of a file and return it as a string

\tArgs:
\t\tfile_path (str): The path to the file
\t\tencoding  (str): The encoding to use when opening the file (default: "utf-8")
\tReturns:
\t\tstr: The content of the file
\t'''
def replace_tilde(path: str) -> str:
    ''' Replace the "~" by the user\'s home directory

\tArgs:
\t\tpath (str): The path to replace the "~" by the user\'s home directory
\tReturns:
\t\tstr: The path with the "~" replaced by the user\'s home directory
\tExamples:

\t\t.. code-block:: python

\t\t\t> replace_tilde("~/Documents/test.txt")
\t\t\t\'/home/user/Documents/test.txt\'
\t'''
def clean_path(file_path: str, trailing_slash: bool = True) -> str:
    ''' Clean the path by replacing backslashes with forward slashes and simplifying the path

\tArgs:
\t\tfile_path (str): The path to clean
\t\ttrailing_slash (bool): Whether to keep the trailing slash, ex: "test/" -> "test/"
\tReturns:
\t\tstr: The cleaned path
\tExamples:
\t\t>>> clean_path("C:\\\\Users\\\\Stoupy\\\\Documents\\\\test.txt")
\t\t\'C:/Users/Stoupy/Documents/test.txt\'

\t\t>>> clean_path("Some Folder////")
\t\t\'Some Folder/\'

\t\t>>> clean_path("test/uwu/1/../../")
\t\t\'test/\'

\t\t>>> clean_path("some/./folder/../")
\t\t\'some/\'

\t\t>>> clean_path("folder1/folder2/../../folder3")
\t\t\'folder3\'

\t\t>>> clean_path("./test/./folder/")
\t\t\'test/folder/\'

\t\t>>> clean_path("C:/folder1\\\\folder2")
\t\t\'C:/folder1/folder2\'
\t'''




# Imports
from collections.abc import Callable
from typing import TYPE_CHECKING, Any, cast

if TYPE_CHECKING:
	import numpy as np
	from numpy.typing import NDArray
	from PIL import Image

# Functions
def auto_crop[T: "Image.Image | NDArray[np.number]"](
	image: T,
	mask: "NDArray[np.bool_] | None" = None,
	threshold: int | float | Callable[["NDArray[np.number]"], int | float] | None = None,
	return_type: type[T] | str = "same",
	contiguous: bool = True,
) -> Any:
	""" Automatically crop an image to remove zero or uniform regions.

	This function crops the image to keep only the region where pixels are non-zero
	(or above a threshold). It can work with a mask or directly analyze the image.

	Args:
		image       (Image.Image | NDArray):	  The image to crop.
		mask        (NDArray[bool] | None):       Optional binary mask indicating regions to keep.
		threshold   (int | float | Callable):     Threshold value or function (default: np.min).
		return_type (type | str):                 Type of the return value (Image.Image, NDArray[np.number], or "same" to match input type).
		contiguous  (bool):                       If True (default), crop to bounding box. If False, remove entire rows/columns with no content.
	Returns:
		Image.Image | NDArray[np.number]: The cropped image.

	Examples:
		>>> # Test with numpy array with zeros on edges
		>>> import numpy as np
		>>> array = np.zeros((100, 100, 3), dtype=np.uint8)
		>>> array[20:80, 30:70] = 255  # White rectangle in center
		>>> cropped = auto_crop(array, return_type=np.ndarray)
		>>> cropped.shape
		(60, 40, 3)

		>>> # Test with custom mask
		>>> mask = np.zeros((100, 100), dtype=bool)
		>>> mask[10:90, 10:90] = True
		>>> cropped_with_mask = auto_crop(array, mask=mask, return_type=np.ndarray)
		>>> cropped_with_mask.shape
		(80, 80, 3)

		>>> # Test with PIL Image
		>>> from PIL import Image
		>>> pil_image = Image.new('RGB', (100, 100), (0, 0, 0))
		>>> from PIL import ImageDraw
		>>> draw = ImageDraw.Draw(pil_image)
		>>> draw.rectangle([25, 25, 75, 75], fill=(255, 255, 255))
		>>> cropped_pil = auto_crop(pil_image)
		>>> cropped_pil.size
		(51, 51)

		>>> # Test with threshold
		>>> array_gray = np.ones((100, 100), dtype=np.uint8) * 10
		>>> array_gray[20:80, 30:70] = 255
		>>> cropped_threshold = auto_crop(array_gray, threshold=50, return_type=np.ndarray)
		>>> cropped_threshold.shape
		(60, 40)

		>>> # Test with callable threshold (using lambda to avoid min value)
		>>> array_gray2 = np.ones((100, 100), dtype=np.uint8) * 10
		>>> array_gray2[20:80, 30:70] = 255
		>>> cropped_max = auto_crop(array_gray2, threshold=lambda x: 50, return_type=np.ndarray)
		>>> cropped_max.shape
		(60, 40)

	>>> # Test with non-contiguous crop
	>>> array_sparse = np.zeros((100, 100, 3), dtype=np.uint8)
	>>> array_sparse[10, 10] = 255
	>>> array_sparse[50, 50] = 255
	>>> array_sparse[90, 90] = 255
	>>> cropped_contiguous = auto_crop(array_sparse, contiguous=True, return_type=np.ndarray)
	>>> cropped_contiguous.shape  # Bounding box from (10,10) to (90,90)
	(81, 81, 3)
	>>> cropped_non_contiguous = auto_crop(array_sparse, contiguous=False, return_type=np.ndarray)
	>>> cropped_non_contiguous.shape  # Only rows/cols 10, 50, 90
	(3, 3, 3)

	>>> # Test with 3D crop on depth dimension
	>>> array_3d = np.zeros((50, 50, 10), dtype=np.uint8)
	>>> array_3d[10:40, 10:40, 2:8] = 255  # Content only in depth slices 2-7
	>>> cropped_3d = auto_crop(array_3d, contiguous=True, return_type=np.ndarray)
	>>> cropped_3d.shape  # Should crop all 3 dimensions
	(30, 30, 6)
	"""
	# Imports
	import numpy as np
	from PIL import Image

	# Convert to numpy array and store original type
	original_was_pil: bool = isinstance(image, Image.Image)
	image_array: NDArray[np.number] = np.array(image) if original_was_pil else image # type: ignore

	# Create mask if not provided
	if mask is None:
		if threshold is None:
			threshold = cast(Callable[["NDArray[np.number]"], int | float], np.min)
		threshold_value: int | float = threshold(image_array) if callable(threshold) else threshold
		# Create a 2D mask for both 2D and 3D arrays
		if image_array.ndim == 2:
			mask = image_array > threshold_value
		else:  # 3D array
			mask = np.any(image_array > threshold_value, axis=2)

	# Find rows, columns, and depth with content
	rows_with_content: NDArray[np.bool_] = np.any(mask, axis=1)
	cols_with_content: NDArray[np.bool_] = np.any(mask, axis=0)

	# For 3D arrays, also find which depth slices have content
	depth_with_content: NDArray[np.bool_] | None = None
	if image_array.ndim == 3:
		# Create a 1D mask for depth dimension
		depth_with_content = np.any(image_array > (threshold(image_array) if callable(threshold) else threshold if threshold is not None else np.min(image_array)), axis=(0, 1))

	# Return original if no content found
	if not (np.any(rows_with_content) and np.any(cols_with_content)):
		return image_array if return_type != Image.Image else (image if original_was_pil else Image.fromarray(image_array))

	# Crop based on contiguous parameter
	if contiguous:
		row_idx, col_idx = np.where(rows_with_content)[0], np.where(cols_with_content)[0]
		if image_array.ndim == 3 and depth_with_content is not None and np.any(depth_with_content):
			depth_idx = np.where(depth_with_content)[0]
			cropped_array: NDArray[np.number] = image_array[row_idx[0]:row_idx[-1]+1, col_idx[0]:col_idx[-1]+1, depth_idx[0]:depth_idx[-1]+1]
		else:
			cropped_array: NDArray[np.number] = image_array[row_idx[0]:row_idx[-1]+1, col_idx[0]:col_idx[-1]+1]
	else:
		if image_array.ndim == 3 and depth_with_content is not None:
			# np.ix_ needs index arrays, not boolean arrays
			row_indices = np.where(rows_with_content)[0]
			col_indices = np.where(cols_with_content)[0]
			depth_indices = np.where(depth_with_content)[0]
			ix = np.ix_(row_indices, col_indices, depth_indices)
		else:
			row_indices = np.where(rows_with_content)[0]
			col_indices = np.where(cols_with_content)[0]
			ix = np.ix_(row_indices, col_indices)
		cropped_array = image_array[ix]

	# Return in requested format
	if return_type == "same":
		return Image.fromarray(cropped_array) if original_was_pil else cropped_array
	return cropped_array if return_type != Image.Image else Image.fromarray(cropped_array)




import zarr
from numpy.typing import NDArray as NDArray
from typing import Any

def array_to_disk(data: NDArray[Any] | zarr.Array[Any], delete_input: bool = True, more_data: NDArray[Any] | zarr.Array[Any] | None = None) -> tuple['zarr.Array[Any]', str, int]:
    """ Easily handle large numpy arrays on disk using zarr for efficient storage and access.

\tZarr provides a simpler and more efficient alternative to np.memmap with better compression
\tand chunking capabilities.

\tArgs:
\t\tdata\t\t\t(NDArray | zarr.Array):\tThe data to save/load as a zarr array
\t\tdelete_input\t(bool):\tWhether to delete the input data after creating the zarr array
\t\tmore_data\t\t(NDArray | zarr.Array | None): Additional data to append to the zarr array
\tReturns:
\t\ttuple[zarr.Array, str, int]: The zarr array, the directory path, and the total size in bytes

\tExamples:
\t\t>>> import numpy as np
\t\t>>> data = np.random.rand(1000, 1000)
\t\t>>> zarr_array = array_to_disk(data)[0]
\t\t>>> zarr_array.shape
\t\t(1000, 1000)

\t\t>>> more_data = np.random.rand(500, 1000)
\t\t>>> longer_array, dir_path, total_size = array_to_disk(zarr_array, more_data=more_data)
\t"""




# Imports
from typing import TYPE_CHECKING, Any

# Lazy imports for typing
if TYPE_CHECKING:
	import polars as pl

# Functions
def upsert_in_dataframe(
	df: "pl.DataFrame",
	new_entry: dict[str, Any],
	primary_keys: list[str] | dict[str, Any] | None = None
) -> "pl.DataFrame":
	""" Insert or update a row in the Polars DataFrame based on primary keys.

	Args:
		df				(pl.DataFrame):		The Polars DataFrame to update.
		new_entry		(dict[str, Any]):	The new entry to insert or update.
		primary_keys	(list[str] | dict[str, Any] | None):	The primary keys to identify the row (for updates).
	Returns:
		pl.DataFrame: The updated Polars DataFrame.
	Examples:
		>>> import polars as pl
		>>> df = pl.DataFrame({"id": [1, 2], "value": ["a", "b"]})
		>>> new_entry = {"id": 2, "value": "updated"}
		>>> updated_df = upsert_in_dataframe(df, new_entry, primary_keys=["id"])
		>>> print(updated_df)
		shape: (2, 2)
		┌─────┬─────────┐
		│ id  ┆ value   │
		│ --- ┆ ---     │
		│ i64 ┆ str     │
		╞═════╪═════════╡
		│ 1   ┆ a       │
		│ 2   ┆ updated │
		└─────┴─────────┘

		>>> new_entry = {"id": 3, "value": "new"}
		>>> updated_df = upsert_in_dataframe(updated_df, new_entry, primary_keys=["id"])
		>>> print(updated_df)
		shape: (3, 2)
		┌─────┬─────────┐
		│ id  ┆ value   │
		│ --- ┆ ---     │
		│ i64 ┆ str     │
		╞═════╪═════════╡
		│ 1   ┆ a       │
		│ 2   ┆ updated │
		│ 3   ┆ new     │
		└─────┴─────────┘
	"""
	# Imports
	import polars as pl

	# Create new DataFrame if file doesn't exist or is invalid
	if df.is_empty():
		return pl.DataFrame([new_entry])

	# If no primary keys provided, return DataFrame with new entry appended
	if not primary_keys:
		new_row_df = pl.DataFrame([new_entry])
		return pl.concat([df, new_row_df], how="diagonal_relaxed")

	# If primary keys are provided as a list, convert to dict with values from new_entry
	if isinstance(primary_keys, list):
		primary_keys = {key: new_entry[key] for key in primary_keys if key in new_entry}

	# Build mask based on primary keys
	mask: pl.Expr = pl.lit(True)
	for key, value in primary_keys.items():
		if key in df.columns:
			mask = mask & (df[key] == value)
		else:
			# Primary key column doesn't exist, so no match possible
			mask = pl.lit(False)
			break

	# Insert or update row based on primary keys
	if df.select(mask).to_series().any():
		# Update existing row
		for key, value in new_entry.items():
			if key in df.columns:
				df = df.with_columns(pl.when(mask).then(pl.lit(value)).otherwise(pl.col(key)).alias(key))
			else:
				# Add new column if it doesn't exist
				df = df.with_columns(pl.when(mask).then(pl.lit(value)).otherwise(None).alias(key))
		return df
	else:
		# Insert new row
		new_row_df = pl.DataFrame([new_entry])
		return pl.concat([df, new_row_df], how="diagonal_relaxed")




from .decorators import LogLevels as LogLevels, deprecated as deprecated
from .io.csv import csv_dump as csv_dump, csv_load as csv_load
from .io.json import json_dump as json_dump, json_load as json_load
from typing import Any

def super_csv_dump(*args: Any, **kwargs: Any) -> Any:
    """ Deprecated function, use :py:func:`~stouputils.io.csv.csv_dump` instead. """
def super_csv_load(*args: Any, **kwargs: Any) -> Any:
    """ Deprecated function, use :py:func:`~stouputils.io.csv.csv_load` instead. """
def super_json_dump(*args: Any, **kwargs: Any) -> Any:
    """ Deprecated function, use :py:func:`~stouputils.io.json.json_dump` instead. """
def super_json_load(*args: Any, **kwargs: Any) -> Any:
    """ Deprecated function, use :py:func:`~stouputils.io.json.json_load` instead. """




# Imports
from collections.abc import Callable
from functools import wraps
from pickle import dumps as pickle_dumps
from typing import Any, Literal, overload

from .common import get_wrapper_name, set_wrapper_name


# Easy cache function with parameter caching method
@overload
def simple_cache[T](
	func: Callable[..., T],
	*,
	method: Literal["str", "pickle"] = "str"
) -> Callable[..., T]: ...

@overload
def simple_cache[T](
	func: None = None,
	*,
	method: Literal["str", "pickle"] = "str"
) -> Callable[[Callable[..., T]], Callable[..., T]]: ...

def simple_cache[T](
	func: Callable[..., T] | None = None,
	*,
	method: Literal["str", "pickle"] = "str"
) -> Callable[..., T] | Callable[[Callable[..., T]], Callable[..., T]]:
	""" Decorator that caches the result of a function based on its arguments.

	The str method is often faster than the pickle method (by a little) but not as accurate with complex objects.

	Args:
		func   (Callable[..., T] | None): Function to cache
		method (Literal["str", "pickle"]):  The method to use for caching.

	Examples:
		>>> @simple_cache
		... def test1(a: int, b: int) -> int:
		...     return a + b

		>>> @simple_cache(method="str")
		... def test2(a: int, b: int) -> int:
		...     return a + b
		>>> test2(1, 2)
		3
		>>> test2(1, 2)
		3
		>>> test2(3, 4)
		7

		>>> @simple_cache
		... def factorial(n: int) -> int:
		...     return n * factorial(n - 1) if n else 1
		>>> factorial(10)   # no previously cached result, makes 11 recursive calls
		3628800
		>>> factorial(5)    # no new calls, just returns the cached result
		120
		>>> factorial(12)   # two new recursive calls, factorial(10) is cached
		479001600
	"""
	def decorator(func: Callable[..., T]) -> Callable[..., T]:
		# Create the cache dict
		cache_dict: dict[Any, Any] = {}

		# Create the wrapper
		@wraps(func)
		def wrapper(*args: tuple[Any, ...], **kwargs: dict[str, Any]) -> Any:

			# Get the hashed key
			if method == "str":
				hashed = str(args) + str(kwargs)
			elif method == "pickle":
				hashed = pickle_dumps((args, kwargs))
			else:
				raise ValueError("Invalid caching method. Supported methods are 'str' and 'pickle'.")

			# If the key is in the cache, return it
			if hashed in cache_dict:
				return cache_dict[hashed]

			# Else, call the function and add the result to the cache
			else:
				result: Any = func(*args, **kwargs)
				cache_dict[hashed] = result
				return result

		# Return the wrapper
		set_wrapper_name(wrapper, get_wrapper_name("stouputils.decorators.simple_cache", func))
		return wrapper

	# Handle both @simple_cache and @simple_cache(method=...)
	if func is None:
		return decorator
	return decorator(func)




""" Common functions used by the Linux and Windows installers modules. """
# Imports
from typing import Literal

from ..print.message import warning


# Functions
def prompt_for_path(prompt_message: str, default_path: str) -> str:
	""" Prompt the user to override a default path.

	Args:
		prompt_message (str): The message to display to the user.
		default_path   (str): The default path to suggest.

	Returns:
		str: The path entered by the user, or the default path if they pressed Enter.
	"""
	warning(f"{prompt_message}\nPress Enter to use this path, or type a new path to override it: ")
	return input() or default_path


def ask_install_type(ask_global: int, default_local_path: str, default_global_path: str | None) -> Literal["g", "l"]:
	""" Determine the installation type (global 'g' or local 'l') based on user input.

	Args:
		ask_global          (int):           0 = ask, 1 = force global, 2 = force local.
		default_local_path  (str):           The default local path.
		default_global_path (str | None):    The default global path (if applicable).

	Returns:
		Literal["g", "l"]: 'g' for global install, 'l' for local install.

	Examples:
		.. code-block:: python

			> # Ask the user while providing default paths
			> install_choice: str = ask_install_type(0, f"{os.getcwd()}/MyProgram", "C:\\Program Files\\MyProgram")
			g

			> # Don't ask, force global
			> install_choice: str = ask_install_type(1, ...)
			g

			> # Don't ask, force local
			> install_choice: str = ask_install_type(2, ...)
			l
	"""
	install_choice: str = ""
	if ask_global == 0:
		if default_global_path:
			global_prompt: str = f"(Globally would target '{default_global_path}')"
		else:
			global_prompt: str = "(Global install not well-defined)"
		warning(
			f"Install globally (requires admin/sudo, suggests adding to PATH) or locally? (G/l):\n"
			f"{global_prompt}, locally would be '{default_local_path}')"
		)
		install_choice = input().lower()
	elif ask_global == 1:
		install_choice = "g"
	elif ask_global == 2:
		install_choice = "l"

	# Default to global unless user explicitly chooses local ('l')
	return 'l' if install_choice == 'l' else 'g'




from collections.abc import Callable as Callable, Iterable
from typing import Literal

def unique_list[T](list_to_clean: Iterable[T], method: Literal['id', 'hash', 'str'] = 'str') -> list[T]:
    ''' Remove duplicates from the list while keeping the order using ids, hash, or str

\tArgs:
\t\tlist_to_clean\t(Iterable[T]):\t\t\t\t\tThe list to clean
\t\tmethod\t\t\t(Literal["id", "hash", "str"]):\tThe method to use to identify duplicates
\tReturns:
\t\tlist[T]: The cleaned list

\tExamples:
\t\t>>> unique_list([1, 2, 3, 2, 1], method="id")
\t\t[1, 2, 3]

\t\t>>> s1 = {1, 2, 3}
\t\t>>> s2 = {2, 3, 4}
\t\t>>> s3 = {1, 2, 3}
\t\t>>> unique_list([s1, s2, s1, s1, s3, s2, s3], method="id")
\t\t[{1, 2, 3}, {2, 3, 4}, {1, 2, 3}]

\t\t>>> s1 = {1, 2, 3}
\t\t>>> s2 = {2, 3, 4}
\t\t>>> s3 = {1, 2, 3}
\t\t>>> unique_list([s1, s2, s1, s1, s3, s2, s3], method="str")
\t\t[{1, 2, 3}, {2, 3, 4}]
\t'''
def at_least_n[T](iterable: Iterable[T], predicate: Callable[[T], bool], n: int) -> bool:
    """ Return True if at least n elements in iterable satisfy predicate.
\tIt's like the built-in any() but for at least n matches.

\tStops iterating as soon as n matches are found (short-circuit evaluation).

\tArgs:
\t\titerable\t(Iterable[T]):\t\t\tThe iterable to check.
\t\tpredicate\t(Callable[[T], bool]):\tThe predicate to apply to items.
\t\tn\t\t\t(int):\t\t\t\t\tMinimum number of matches required.

\tReturns:
\t\tbool: True if at least n elements satisfy predicate, otherwise False.

\tExamples:
\t\t>>> at_least_n([1, 2, 3, 4, *[i for i in range(5, int(1e5))]], lambda x: x % 2 == 0, 2)
\t\tTrue
\t\t>>> at_least_n([1, 3, 5, 7], lambda x: x % 2 == 0, 1)
\t\tFalse
\t"""



from ..decorators import handle_error as handle_error, measure_time as measure_time
from ..io.path import clean_path as clean_path
from ..print.message import info as info, warning as warning
from .hash import get_file_hash as get_file_hash
from .retrieve import get_all_previous_backups as get_all_previous_backups, is_file_in_any_previous_backup as is_file_in_any_previous_backup

@handle_error
def create_delta_backup(source_path: str, destination_folder: str, exclude_patterns: list[str] | None = None) -> None:
    ''' Creates a ZIP delta backup, saving only modified or new files while tracking deleted files.

\tArgs:
\t\tsource_path (str): Path to the source file or directory to back up
\t\tdestination_folder (str): Path to the folder where the backup will be saved
\t\texclude_patterns (list[str] | None): List of glob patterns to exclude from backup
\tExamples:

\t.. code-block:: python

\t\t> create_delta_backup("/path/to/source", "/path/to/backups", exclude_patterns=["libraries/*", "cache/*"])
\t\t[INFO HH:MM:SS] Creating ZIP backup
\t\t[INFO HH:MM:SS] Backup created: \'/path/to/backups/backup_2025_02_18-10_00_00.zip\'
\t'''



from ..decorators import handle_error as handle_error
from ..io.path import super_open as super_open
from ..print.message import info as info, progress as progress, warning as warning
from .cd_utils import clean_version as clean_version, format_changelog as format_changelog, version_to_float as version_to_float
from collections.abc import Callable as Callable

def run_git_command(args: list[str], cwd: str | None = None) -> str:
    """ Run a git command and return the output.

\tArgs:
\t\targs\t(list[str]):\tGit command arguments (without 'git' prefix)
\t\tcwd\t\t(str | None):\tWorking directory for the command
\tReturns:
\t\tstr:\tCommand output (stdout)
\tRaises:
\t\t:py:exc:`RuntimeError`: If the git command fails
\t"""
def get_local_tags(cwd: str | None = None) -> list[tuple[str, str]]:
    """ Get all tags from the local git repository, sorted by version.

\tArgs:
\t\tcwd\t(str | None):\tWorking directory for the git command
\tReturns:
\t\tlist[tuple[str, str]]:\tList of (tag_name, commit_sha) tuples, sorted by version (newest first)
\t"""
def get_latest_tag(cwd: str | None = None, exclude_version: str | None = None) -> tuple[str, str] | tuple[None, None]:
    """ Get the latest tag from the local git repository.

\tArgs:
\t\tcwd\t\t\t\t(str | None):\tWorking directory for the git command
\t\texclude_version\t(str | None):\tVersion to exclude from the search
\tReturns:
\t\ttuple[str, str] | tuple[None, None]:\t(tag_name, commit_sha) or (None, None) if no tags exist
\t"""
def get_commits_since_tag(tag: str, cwd: str | None = None) -> list[tuple[str, str]]:
    """ Get all commits since a specific tag.

\tArgs:
\t\ttag\t(str):\t\t\tTag name to start from (exclusive)
\t\tcwd\t(str | None):\tWorking directory for the git command
\tReturns:
\t\tlist[tuple[str, str]]:\tList of (sha, message) tuples
\t"""
def get_commits_since_date(date_str: str, cwd: str | None = None) -> list[tuple[str, str]]:
    """ Get all commits since a specific date.

\tArgs:
\t\tdate_str\t(str):\t\t\tDate string (supports multiple formats via dateutil)
\t\tcwd\t\t\t(str | None):\tWorking directory for the git command
\tReturns:
\t\tlist[tuple[str, str]]:\tList of (sha, message) tuples
\t"""
def get_commits_since_commit(commit_sha: str, cwd: str | None = None) -> list[tuple[str, str]]:
    """ Get all commits since a specific commit (exclusive).

\tArgs:
\t\tcommit_sha\t(str):\t\t\tCommit SHA to start from (this commit is excluded)
\t\tcwd\t\t\t(str | None):\tWorking directory for the git command
\tReturns:
\t\tlist[tuple[str, str]]:\tList of (sha, message) tuples
\t"""
def parse_date_fallback(date_str: str) -> str:
    ''' Parse a date string without dateutil, trying common formats.

\tArgs:
\t\tdate_str\t(str):\tDate string to parse
\tReturns:
\t\tstr:\tISO 8601 formatted date string
\tRaises:
\t\t:py:exc:`ValueError`:\tIf the date cannot be parsed

\t>>> parse_date_fallback("2026/01/15")
\t\'2026-01-15T00:00:00\'
\t>>> parse_date_fallback("2026-01-15")
\t\'2026-01-15T00:00:00\'
\t>>> parse_date_fallback("2026-01-15 14:30:00")
\t\'2026-01-15T14:30:00\'
\t>>> parse_date_fallback("2026-01-15T14:30:00")
\t\'2026-01-15T14:30:00\'
\t'''
def parse_commit_log(output: str) -> list[tuple[str, str]]:
    """ Parse git log output into a list of (sha, message) tuples.

\tArgs:
\t\toutput\t(str):\tOutput from git log command
\tReturns:
\t\tlist[tuple[str, str]]:\tList of (sha, full_message) tuples
\t"""
def get_remotes(cwd: str | None = None) -> dict[str, str]:
    """ Get all git remotes and their URLs.

\tArgs:
\t\tcwd\t(str | None):\tWorking directory for the git command
\tReturns:
\t\tdict[str, str]:\tDictionary mapping remote names to their push URLs
\t"""
def parse_remote_url(remote_url: str) -> tuple[str, str, str] | None:
    ''' Parse a git remote URL to extract hosting info.

\tSupports:
\t- SSH format: git@github.com:user/repo.git
\t- SSH format: git@gitlab.example.com:group/repo.git
\t- HTTPS format: https://github.com/user/repo.git
\t- HTTPS format: https://gitlab.example.com/group/repo.git

\tArgs:
\t\tremote_url\t(str):\tGit remote URL
\tReturns:
\t\ttuple[str, str, str] | None:\t(host_type, base_url, repo_path) or None if cannot parse
\t\t\thost_type:\thost_type: "github", "gitlab", or "unknown"
\t\t\tbase_url:\tBase URL for the repository (e.g., "https://github.com/user/repo")
\t\t\trepo_path:\tRepository path (e.g., "user/repo")

\t>>> parse_remote_url("git@github.com:Stoupy51/stouputils.git")
\t(\'github\', \'https://github.com/Stoupy51/stouputils\', \'Stoupy51/stouputils\')
\t>>> parse_remote_url("https://github.com/Stoupy51/stouputils.git")
\t(\'github\', \'https://github.com/Stoupy51/stouputils\', \'Stoupy51/stouputils\')
\t>>> parse_remote_url("git@gitlab.example.com:group/project.git")
\t(\'gitlab\', \'https://gitlab.example.com/group/project\', \'group/project\')
\t>>> parse_remote_url("https://gitlab.company.com/team/repo.git")
\t(\'gitlab\', \'https://gitlab.company.com/team/repo\', \'team/repo\')
\t>>> parse_remote_url("git@custom-server.com:user/repo.git")
\t(\'gitlab\', \'https://custom-server.com/user/repo\', \'user/repo\')
\t>>> parse_remote_url("invalid-url") is None
\tTrue
\t'''
def detect_host_type(host: str) -> str:
    ''' Detect the type of git hosting service from hostname.

\tArgs:
\t\thost (str):\tHostname (e.g., "github.com", "gitlab.example.com")
\tReturns:
\t\tstr: "github", "gitlab", or "unknown"

\t>>> detect_host_type("github.com")
\t\'github\'
\t>>> detect_host_type("gitlab.com")
\t\'gitlab\'
\t>>> detect_host_type("gitlab.example.com")
\t\'gitlab\'
\t>>> detect_host_type("custom-server.com")
\t\'gitlab\'
\t>>> detect_host_type("my-github-mirror.org")
\t\'github\'
\t'''
def create_url_formatter(remote_url: str) -> tuple[Callable[[str], str], Callable[[str, str], str]] | None:
    """ Create URL formatter functions for a git remote.

\tArgs:
\t\tremote_url\t(str):\tGit remote URL
\tReturns:
\t\ttuple[Callable, Callable] | None:\t(commit_url_formatter, compare_url_formatter) or None
\t"""
def generate_local_changelog(mode: str = 'tag', value: str | None = None, remote: str | None = None, cwd: str | None = None) -> str:
    ''' Generate a changelog from local git history.

\tArgs:
\t\tmode\t(str):\t\t\tMode for selecting commits - "tag", "date", or "commit"
\t\tvalue\t(str | None):\tValue for the mode (tag name, date, or commit SHA).
\t\t\tIf None and mode is "tag", uses the latest tag.
\t\tremote\t(str | None):\tRemote name to use for commit URLs. If None, no URLs are generated.
\t\tcwd\t\t(str | None):\tWorking directory for git commands
\tReturns:
\t\tstr:\tGenerated changelog in Markdown format
\t'''
def changelog_cli() -> None:
    """ CLI interface for generating changelogs from local git history.

\tUsage:
\t\tstouputils changelog [tag|date|commit] [value] [--remote <remote>] [-o <file>]

\tExamples:
\t\tstouputils changelog                          # Uses latest tag (default)
\t\tstouputils changelog tag v1.9.0               # All commits since tag v1.9.0
\t\tstouputils changelog date 2026/01/05          # All commits since date
\t\tstouputils changelog commit 847b27e           # All commits since commit
\t\tstouputils changelog --remote origin          # Use origin remote for commit URLs
\t\tstouputils changelog -o CHANGELOG.md          # Output to file
\t"""



""" Installer module for stouputils.

Provides functions for platform-agnostic installation tasks by dispatching
to platform-specific implementations (Windows, Linux/macOS).

It handles getting installation paths, adding programs to the PATH environment variable,
and installing programs from local zip files or URLs.

Main functions:

- :py:func:`~downloader.check_executable`: Check if an executable exists and offer to download it
- :py:func:`~downloader.download_executable`: Download and install a program from GitHub releases
- :py:func:`~main.install_program`: Install a program from a local archive or URL
- :py:func:`~main.get_install_path`: Get the appropriate installation path for the current platform
- :py:func:`~main.add_to_path`: Add a directory to the system PATH

Example usage from the upscaler application:

.. code-block:: python

    from stouputils.installer import check_executable

    # Define download URLs for different platforms
    WAIFU2X_RELEASES = {
        "Windows": "https://github.com/nihui/waifu2x-ncnn-vulkan/releases/download/20220728/waifu2x-ncnn-vulkan-20220728-windows.zip",
        "Linux": "https://github.com/nihui/waifu2x-ncnn-vulkan/releases/download/20220728/waifu2x-ncnn-vulkan-20220728-ubuntu.zip",
        "Darwin": "https://github.com/nihui/waifu2x-ncnn-vulkan/releases/download/20220728/waifu2x-ncnn-vulkan-20220728-macos.zip"
    }

    # Check if the executable exists, download if needed
    check_executable(
        "waifu2x-ncnn-vulkan",
        "waifu2x-ncnn-vulkan",  # Help text to verify the executable
        WAIFU2X_RELEASES
    )

    # For executables in a subdirectory (e.g., FFmpeg in bin/)
    FFMPEG_RELEASES = {
        "Windows": "https://github.com/BtbN/FFmpeg-Builds/releases/download/latest/ffmpeg-master-latest-win64-gpl.zip",
        "Linux": "https://github.com/BtbN/FFmpeg-Builds/releases/download/latest/ffmpeg-master-latest-linux64-gpl.tar.xz",
    }
    check_executable(
        "ffmpeg",
        "ffmpeg version",
        FFMPEG_RELEASES,
        append_to_path="bin"  # Executables are in the bin/ subdirectory
    )

Manual installation example:

.. code-block:: python

    from stouputils.installer import install_program, get_install_path, add_to_path

    # Get the appropriate installation path
    install_path = get_install_path(
        "my-program",
        ask_global=0,  # Ask user: 0=ask, 1=force global, 2=force local
        add_path=True,
        append_to_path="bin"
    )

    # Install from a URL
    success = install_program(
        "https://example.com/program.zip",
        install_path=install_path,
        program_name="my-program",
        add_path=True,
        append_to_path="bin"
    )

    # Or install from a local file
    success = install_program(
        "/path/to/program.zip",
        install_path=install_path,
        program_name="my-program"
    )

    # Manually add to PATH if needed
    if success:
        add_to_path(f"{install_path}/bin")
"""

# Imports
from .common import *
from .downloader import *
from .linux import *
from .main import *
from .windows import *




from collections.abc import Callable, Iterable, Mapping, MutableMapping
from typing import Any

type JsonDict = dict[str, Any]
type JsonList = list[Any]
type JsonMap = Mapping[str, Any]
type JsonMutMap = MutableMapping[str, Any]
type IterAny = Iterable[Any]
type CallableAny = Callable[..., Any]
def convert_to_serializable(obj: Any) -> Any:
    """ Recursively convert objects to JSON-serializable forms.

\tObjects with a `to_dict()` or `asdict()` method are converted to their dictionary representation.
\tDictionaries and lists are recursively processed.

\tCan also be used to convert nested structures containing custom objects,
\tsuch as defaultdict, dataclasses, or other user-defined types.

\tArgs:
\t\tobj (Any): The object to convert
\tReturns:
\t\tAny: The JSON-serializable version of the object
\tExamples:
\t\t>>> from typing import defaultdict
\t\t>>> my_dict = defaultdict(lambda: defaultdict(int))
\t\t>>> my_dict['a']['b'] += 6
\t\t>>> my_dict['c']['d'] = 4
\t\t>>> my_dict['a']
\t\tdefaultdict(<class 'int'>, {'b': 6})
\t\t>>> my_dict['c']
\t\tdefaultdict(<class 'int'>, {'d': 4})
\t\t>>> convert_to_serializable(my_dict)
\t\t{'a': {'b': 6}, 'c': {'d': 4}}

\t\t>>> from dataclasses import dataclass
\t\t>>> @dataclass
\t\t... class Point:
\t\t...     x: int
\t\t...     y: int
\t\t>>> convert_to_serializable(Point(3, 4))
\t\t{'x': 3, 'y': 4}
\t"""




# Imports
from __future__ import annotations

from typing import Any

from .common import AbstractBothContextManager


# Context manager to temporarily set multiprocessing start method
class SetMPStartMethod(AbstractBothContextManager["SetMPStartMethod"]):
	""" Context manager to temporarily set multiprocessing start method.

	This context manager allows you to temporarily change the multiprocessing start method
	and automatically restores the original method when exiting the context.

	Args:
		start_method (str): The start method to use: "spawn", "fork", or "forkserver"

	Examples:
		.. code-block:: python

			> import multiprocessing as mp
			> import stouputils as stp
			> # Temporarily use spawn method
			> with stp.SetMPStartMethod("spawn"):
			> ...     # Your multiprocessing code here
			> ...     pass

			> # Original method is automatically restored
	"""
	def __init__(self, start_method: str | None) -> None:
		self.start_method: str | None = start_method
		""" The start method to use """
		self.old_method: str | None = None
		""" The original start method to restore """

	def __enter__(self) -> SetMPStartMethod:
		""" Enter context manager which sets the start method """
		if self.start_method is None:
			return self
		import multiprocessing as mp

		self.old_method = mp.get_start_method(allow_none=True)
		if self.old_method != self.start_method:
			mp.set_start_method(self.start_method, force=True)
		return self

	def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
		""" Exit context manager which restores the original start method """
		if self.start_method is None:
			return
		import multiprocessing as mp

		if self.old_method != self.start_method:
			mp.set_start_method(self.old_method, force=True)

	async def __aenter__(self) -> SetMPStartMethod:
		""" Enter async context manager which sets the start method """
		return self.__enter__()

	async def __aexit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
		""" Exit async context manager which restores the original start method """
		self.__exit__(exc_type, exc_val, exc_tb)





# Imports
import datetime
import fnmatch
import os
import zipfile

from ..config import StouputilsConfig as Cfg
from ..decorators import handle_error, measure_time
from ..io.path import clean_path
from ..print.message import info, warning
from .hash import get_file_hash
from .retrieve import get_all_previous_backups, is_file_in_any_previous_backup


# Main backup function that creates a delta backup (only changed files)
@measure_time(message="Creating ZIP backup")
@handle_error
def create_delta_backup(source_path: str, destination_folder: str, exclude_patterns: list[str] | None = None) -> None:
	""" Creates a ZIP delta backup, saving only modified or new files while tracking deleted files.

	Args:
		source_path (str): Path to the source file or directory to back up
		destination_folder (str): Path to the folder where the backup will be saved
		exclude_patterns (list[str] | None): List of glob patterns to exclude from backup
	Examples:

	.. code-block:: python

		> create_delta_backup("/path/to/source", "/path/to/backups", exclude_patterns=["libraries/*", "cache/*"])
		[INFO HH:MM:SS] Creating ZIP backup
		[INFO HH:MM:SS] Backup created: '/path/to/backups/backup_2025_02_18-10_00_00.zip'
	"""
	source_path = clean_path(os.path.abspath(source_path))
	destination_folder = clean_path(os.path.abspath(destination_folder))

	# Setup backup paths and create destination folder
	base_name: str = os.path.basename(source_path.rstrip(os.sep)) or "backup"
	backup_folder: str = clean_path(os.path.join(destination_folder, base_name))
	os.makedirs(backup_folder, exist_ok=True)

	# Get previous backups and track all files
	previous_backups: dict[str, dict[str, str]] = get_all_previous_backups(backup_folder)
	previous_files: set[str] = {file for backup in previous_backups.values() for file in backup}  # Collect all tracked files

	# Create new backup filename with timestamp
	timestamp: str = datetime.datetime.now().strftime("%Y_%m_%d-%H_%M_%S")
	zip_filename: str = f"{timestamp}.zip"
	destination_zip: str = clean_path(os.path.join(backup_folder, zip_filename))

	# Create the ZIP file early to write files as we process them
	with zipfile.ZipFile(destination_zip, "w", compression=zipfile.ZIP_DEFLATED, compresslevel=9) as zipf:
		deleted_files: set[str] = set()
		has_changes: bool = False

		# Process files one by one to avoid memory issues
		if os.path.isdir(source_path):
			for root, _, files in os.walk(source_path):
				for file in files:
					full_path: str = clean_path(os.path.join(root, file))
					arcname: str = clean_path(os.path.relpath(full_path, start=os.path.dirname(source_path)))

					# Skip file if it matches any exclude pattern
					if exclude_patterns and any(fnmatch.fnmatch(arcname, pattern) for pattern in exclude_patterns):
						continue

					file_hash: str | None = get_file_hash(full_path)
					if file_hash is None:
						continue

					# Check if file needs to be backed up
					if not is_file_in_any_previous_backup(arcname, file_hash, previous_backups):
						try:
							zip_info: zipfile.ZipInfo = zipfile.ZipInfo(arcname)
							zip_info.compress_type = zipfile.ZIP_DEFLATED
							zip_info.comment = file_hash.encode()  # Store hash in comment

							# Read and write file in chunks with larger buffer
							with open(full_path, "rb") as f:
								with zipf.open(zip_info, "w", force_zip64=True) as zf:
									while True:
										chunk = f.read(Cfg.CHUNK_SIZE)
										if not chunk:
											break
										zf.write(chunk)
							has_changes = True
						except Exception as e:
							warning(f"Error writing file {full_path} to backup: {e}")

					# Track current files for deletion detection
					if arcname in previous_files:
						previous_files.remove(arcname)
		else:
			arcname: str = clean_path(os.path.basename(source_path))
			file_hash: str | None = get_file_hash(source_path)

			if file_hash is not None and not is_file_in_any_previous_backup(arcname, file_hash, previous_backups):
				try:
					zip_info: zipfile.ZipInfo = zipfile.ZipInfo(arcname)
					zip_info.compress_type = zipfile.ZIP_DEFLATED
					zip_info.comment = file_hash.encode()

					with open(source_path, "rb") as f:
						with zipf.open(zip_info, "w", force_zip64=True) as zf:
							while True:
								chunk = f.read(Cfg.CHUNK_SIZE)
								if not chunk:
									break
								zf.write(chunk)
					has_changes = True
				except Exception as e:
					warning(f"Error writing file {source_path} to backup: {e}")

		# Any remaining files in previous_files were deleted
		deleted_files = previous_files
		if deleted_files:
			zipf.writestr("__deleted_files__.txt", "\n".join(deleted_files), compress_type=zipfile.ZIP_DEFLATED)
			has_changes = True

	# Remove empty backup if no changes
	if not has_changes:
		os.remove(destination_zip)
		info(f"No files to backup, skipping creation of backup '{destination_zip}'")
	else:
		info(f"Backup created: '{destination_zip}'")





# Imports
import time
from collections.abc import Callable
from enum import Enum
from functools import wraps
from traceback import format_exc
from typing import Any, overload

from ..config import StouputilsConfig as Cfg
from ..print.message import error, warning
from .common import get_function_name, get_wrapper_name, set_wrapper_name


# Decorator that handle an error with different log levels
class LogLevels(Enum):
	""" Log level for the errors in the decorator handle_error """
	NONE = 0
	""" Do nothing """
	WARNING = 1
	""" Show as warning """
	WARNING_TRACEBACK = 2
	""" Show as warning with traceback """
	ERROR_TRACEBACK = 3
	""" Show as error with traceback """
	RAISE_EXCEPTION = 4
	""" Raise exception """

@overload
def handle_error[T](
	func: Callable[..., T],
	*,
	exceptions: tuple[type[BaseException], ...] | type[BaseException] = (Exception,),
	message: str = "",
	error_log: LogLevels = LogLevels.WARNING_TRACEBACK,
	sleep_time: float = 0.0
) -> Callable[..., T]: ...

@overload
def handle_error[T](
	func: None = None,
	*,
	exceptions: tuple[type[BaseException], ...] | type[BaseException] = (Exception,),
	message: str = "",
	error_log: LogLevels = LogLevels.WARNING_TRACEBACK,
	sleep_time: float = 0.0
) -> Callable[[Callable[..., T]], Callable[..., T]]: ...

def handle_error[T](
	func: Callable[..., T] | None = None,
	*,
	exceptions: tuple[type[BaseException], ...] | type[BaseException] = (Exception,),
	message: str = "",
	error_log: LogLevels = LogLevels.WARNING_TRACEBACK,
	sleep_time: float = 0.0
) -> Callable[..., T] | Callable[[Callable[..., T]], Callable[..., T]]:
	""" Decorator that handle an error with different log levels.

	Args:
		func        (Callable[..., T] | None):    	Function to decorate
		exceptions	(tuple[type[BaseException]], ...):	Exceptions to handle
		message		(str):								Message to display with the error. (e.g. "Error during something")
		error_log	(LogLevels):						Log level for the errors

			- :attr:`LogLevels.NONE` - None
			- :attr:`LogLevels.WARNING` - Show as warning
			- :attr:`LogLevels.WARNING_TRACEBACK` - Show as warning with traceback
			- :attr:`LogLevels.ERROR_TRACEBACK` - Show as error with traceback
			- :attr:`LogLevels.RAISE_EXCEPTION` - Raise exception

		sleep_time	(float):							Time to sleep after the error (e.g. 0.0 to not sleep, 1.0 to sleep for 1 second)

	Examples:
		>>> @handle_error
		... def might_fail():
		...     raise ValueError("Let's fail")

		>>> @handle_error(error_log=LogLevels.WARNING)
		... def test():
		...     raise ValueError("Let's fail")
		>>> # test()	# [WARNING HH:MM:SS] Error during test: (ValueError) Let's fail
	"""
	# Update error_log if needed
	if Cfg.FORCE_RAISE_EXCEPTION:
		error_log = LogLevels.RAISE_EXCEPTION

	def decorator(func: Callable[..., T]) -> Callable[..., T]:
		if message != "":
			msg: str = f"{message}, "
		else:
			msg: str = message

		@wraps(func)
		def wrapper(*args: tuple[Any, ...], **kwargs: dict[str, Any]) -> Any:
			try:
				return func(*args, **kwargs)
			except exceptions as e:
				if error_log == LogLevels.WARNING:
					warning(f"{msg}Error during {get_function_name(func)}(): ({type(e).__name__}) {e}")
				elif error_log == LogLevels.WARNING_TRACEBACK:
					warning(f"{msg}Error during {get_function_name(func)}():\n{format_exc()}")
				elif error_log == LogLevels.ERROR_TRACEBACK:
					error(f"{msg}Error during {get_function_name(func)}():\n{format_exc()}", exit=True)
				elif error_log == LogLevels.RAISE_EXCEPTION:
					raise e

				# Sleep for the specified time, only if the error_log is not ERROR_TRACEBACK (because it's blocking)
				if sleep_time > 0.0 and error_log != LogLevels.ERROR_TRACEBACK:
					time.sleep(sleep_time)
		set_wrapper_name(wrapper, get_wrapper_name("stouputils.decorators.handle_error", func))
		return wrapper

	# Handle both @handle_error and @handle_error(exceptions=..., message=..., error_log=...)
	if func is None:
		return decorator
	return decorator(func)




from ..decorators import handle_error as handle_error, measure_time as measure_time
from ..io.path import clean_path as clean_path
from ..print.message import info as info, progress as progress, warning as warning
from .cd_utils import clean_version as clean_version, format_changelog as format_changelog, handle_response as handle_response, version_to_float as version_to_float
from typing import Any

PROJECT_ENDPOINT: str

def validate_credentials(credentials: dict[str, dict[str, str]]) -> tuple[str, dict[str, str]]:
    """ Get and validate GitHub credentials

\tArgs:
\t\tcredentials (dict[str, dict[str, str]]):\tCredentials for the GitHub API
\tReturns:
\t\ttuple[str, dict[str, str]]:
\t\t\tstr:\t\t\tOwner (the username of the account to use)

\t\t\tdict[str, str]:\tHeaders (for the requests to the GitHub API)
\t"""
def validate_config(github_config: dict[str, Any]) -> tuple[str, str, str, list[str]]:
    """ Validate GitHub configuration

\tArgs:
\t\tgithub_config (dict[str, str]):\tConfiguration for the GitHub project
\tReturns:
\t\ttuple[str, str, str, list[str]]:
\t\t\tstr: Project name on GitHub

\t\t\tstr: Version of the project

\t\t\tstr: Build folder path containing zip files to upload to the release

\t\t\tlist[str]: List of zip files to upload to the release
\t"""
def handle_existing_tag(owner: str, project_name: str, version: str, headers: dict[str, str]) -> bool:
    """ Check if tag exists and handle deletion if needed

\tArgs:
\t\towner\t\t\t(str):\t\t\t\tGitHub username
\t\tproject_name\t(str):\t\t\t\tName of the GitHub repository
\t\tversion\t\t\t(str):\t\t\t\tVersion to check for existing tag
\t\theaders\t\t\t(dict[str, str]):\tHeaders for GitHub API requests
\tReturns:
\t\tbool: True if the tag was deleted or if it was not found, False otherwise
\t"""
def delete_existing_release(owner: str, project_name: str, version: str, headers: dict[str, str]) -> None:
    """ Delete existing release for a version

\tArgs:
\t\towner\t\t\t(str):\t\t\t\tGitHub username
\t\tproject_name\t(str):\t\t\t\tName of the GitHub repository
\t\tversion\t\t\t(str):\t\t\t\tVersion of the release to delete
\t\theaders\t\t\t(dict[str, str]):\tHeaders for GitHub API requests
\t"""
def delete_existing_tag(tag_url: str, headers: dict[str, str]) -> None:
    """ Delete existing tag

\tArgs:
\t\ttag_url\t(str):\t\t\t\tURL of the tag to delete
\t\theaders\t(dict[str, str]):\tHeaders for GitHub API requests
\t"""
def get_latest_tag(owner: str, project_name: str, version: str, headers: dict[str, str]) -> tuple[str, str] | tuple[None, None]:
    """ Get latest tag information

\tArgs:
\t\towner\t\t\t(str):\t\t\t\tGitHub username
\t\tproject_name\t(str):\t\t\t\tName of the GitHub repository
\t\tversion\t\t\t(str):\t\t\t\tVersion to remove from the list of tags
\t\theaders\t\t\t(dict[str, str]):\tHeaders for GitHub API requests
\tReturns:
\t\tstr|None: SHA of the latest tag commit, None if no tags exist
\t\tstr|None: Version number of the latest tag, None if no tags exist
\t"""
def get_commits_since_tag(owner: str, project_name: str, latest_tag_sha: str | None, headers: dict[str, str]) -> list[dict[str, Any]]:
    """ Get commits since last tag

\tArgs:
\t\towner\t\t\t(str):\t\t\t\tGitHub username
\t\tproject_name\t(str):\t\t\t\tName of the GitHub repository
\t\tlatest_tag_sha\t(str|None):\t\t\tSHA of the latest tag commit
\t\theaders\t\t\t(dict[str, str]):\tHeaders for GitHub API requests
\tReturns:
\t\tlist[dict]: List of commits since the last tag
\t"""
def generate_changelog(commits: list[dict[str, Any]], owner: str, project_name: str, latest_tag_version: str | None, version: str) -> str:
    """ Generate changelog from commits. They must follow the conventional commits convention.

\tConvention format: <type>: <description> or <type>(<sub-category>): <description>

\tArgs:
\t\tcommits\t\t\t\t(list[dict]):\tList of commits to generate changelog from
\t\towner\t\t\t\t(str):\t\t\tGitHub username
\t\tproject_name\t\t(str):\t\t\tName of the GitHub repository
\t\tlatest_tag_version\t(str|None):\t\tVersion number of the latest tag
\t\tversion\t\t\t\t(str):\t\t\tCurrent version being released
\tReturns:
\t\tstr: Generated changelog text
\tSource:
\t\thttps://www.conventionalcommits.org/en/v1.0.0/
\t"""
def create_tag(owner: str, project_name: str, version: str, headers: dict[str, str]) -> None:
    """ Create a new tag

\tArgs:
\t\towner\t\t\t(str):\t\t\t\tGitHub username
\t\tproject_name\t(str):\t\t\t\tName of the GitHub repository
\t\tversion\t\t\t(str):\t\t\t\tVersion for the new tag
\t\theaders\t\t\t(dict[str, str]):\tHeaders for GitHub API requests
\t"""
def create_release(owner: str, project_name: str, version: str, changelog: str, headers: dict[str, str]) -> int:
    """ Create a new release

\tArgs:
\t\towner\t\t\t(str):\t\t\t\tGitHub username
\t\tproject_name\t(str):\t\t\t\tName of the GitHub repository
\t\tversion\t\t\t(str):\t\t\t\tVersion for the new release
\t\tchangelog\t\t(str):\t\t\t\tChangelog text for the release
\t\theaders\t\t\t(dict[str, str]):\tHeaders for GitHub API requests
\tReturns:
\t\tint: ID of the created release
\t"""
def upload_assets(owner: str, project_name: str, release_id: int, build_folder: str, headers: dict[str, str], endswith: list[str]) -> None:
    """ Upload release assets

\tArgs:
\t\towner\t\t\t(str):\t\t\t\tGitHub username
\t\tproject_name\t(str):\t\t\t\tName of the GitHub repository
\t\trelease_id\t\t(int):\t\t\t\tID of the release to upload assets to
\t\tbuild_folder\t(str):\t\t\t\tFolder containing assets to upload
\t\theaders\t\t\t(dict[str, str]):\tHeaders for GitHub API requests
\t\tendswith\t\t(list[str]):\t\tList of files to upload to the release
\t\t\t(every file ending with one of these strings will be uploaded)
\t"""
@handle_error
def upload_to_github(credentials: dict[str, Any], github_config: dict[str, Any]) -> str:
    ''' Upload the project to GitHub using the credentials and the configuration

\tArgs:
\t\tcredentials\t\t(dict[str, Any]):\tCredentials for the GitHub API
\t\tgithub_config\t(dict[str, Any]):\tConfiguration for the GitHub project
\tReturns:
\t\tstr: Generated changelog text
\tExamples:

\t.. code-block:: python

\t\t> upload_to_github(
\t\t\tcredentials={
\t\t\t\t"github": {
\t\t\t\t\t"api_key": "ghp_...",
\t\t\t\t\t"username": "Stoupy"
\t\t\t\t}
\t\t\t},
\t\t\tgithub_config={
\t\t\t\t"project_name": "stouputils",
\t\t\t\t"version": "1.0.0",
\t\t\t\t"build_folder": "build",
\t\t\t\t"endswith": [".zip"]
\t\t\t}
\t\t)
\t'''




# Imports
from __future__ import annotations

from typing import Any

from .common import AbstractBothContextManager


# Context manager that does nothing
class DoNothing(AbstractBothContextManager["DoNothing"]):
	""" Context manager that does nothing.

	This is a no-op context manager that can be used as a placeholder
	or for conditional context management.

	Different from contextlib.nullcontext because it handles args and kwargs,
	along with **async** context management.

	Examples:
		>>> with DoNothing():
		...     print("This will be printed normally")
		This will be printed normally

		>>> # Conditional context management
		>>> some_condition = True
		>>> ctx = DoNothing() if some_condition else Muffle()
		>>> with ctx:
		...     print("May or may not be printed depending on condition")
		May or may not be printed depending on condition
	"""
	def __init__(self, *args: Any, **kwargs: Any) -> None:
		""" No initialization needed, this is a no-op context manager """
		pass

	def __enter__(self) -> DoNothing:
		""" Enter context manager (does nothing) """
		return self

	def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
		""" Exit context manager (does nothing) """
		pass

	async def __aenter__(self) -> DoNothing:
		""" Enter async context manager (does nothing) """
		return self

	async def __aexit__(self, *excinfo: Any) -> None:
		""" Exit async context manager (does nothing) """
		pass

NullContextManager = DoNothing
""" Alias for DoNothing context manager """




""" Installer module for Windows specific functions.

Provides Windows specific implementations for checking administrator privileges,
determining appropriate installation paths (global/local), and modifying
the user's PATH environment variable.
"""
# Imports
import os

from ..decorators import LogLevels, handle_error
from ..io.path import clean_path
from ..print.message import debug, info, warning
from .common import ask_install_type, prompt_for_path


# Functions
@handle_error(message="Failed to add to PATH (Windows)", error_log=LogLevels.WARNING_TRACEBACK)
def add_to_path_windows(install_path: str) -> bool | None:
	""" Add install_path to the User PATH environment variable on Windows.

	Args:
		install_path (str): The path to add to the User PATH environment variable.

	Returns:
		bool | None: True if the path was added to the User PATH environment variable, None otherwise.
	"""
	# Convert install_path to a Windows path if it's not already
	install_path = install_path.replace("/", "\\")
	os.makedirs(install_path, exist_ok=True)

	# Get current user PATH
	import winreg
	with winreg.OpenKey(winreg.HKEY_CURRENT_USER, "Environment", 0, winreg.KEY_READ | winreg.KEY_WRITE) as key:

		# Get the number of values in the registry key
		num_values = winreg.QueryInfoKey(key)[1]

		# Find the index of the 'Path' value
		path_index = -1
		for i in range(num_values):
			if winreg.EnumValue(key, i)[0] == 'Path':
				path_index = i
				break

		# Get the current path value
		current_path: str = winreg.EnumValue(key, path_index)[1]

		# Check if path is already present
		if install_path not in current_path.split(';'):
			new_path: str = f"{current_path};{install_path}"
			winreg.SetValueEx(key, "Path", 0, winreg.REG_EXPAND_SZ, new_path)
			debug(f"Added '{install_path}' to user PATH. Please restart your terminal for changes to take effect.")
		else:
			debug(f"'{install_path}' is already in user PATH.")
	return True


def check_admin_windows() -> bool:
	""" Check if the script is running with administrator privileges on Windows. """
	try:
		import ctypes
		return ctypes.windll.shell32.IsUserAnAdmin() != 0
	except Exception:
		return False


@handle_error(message="Failed to get installation path (Windows)", error_log=LogLevels.ERROR_TRACEBACK)
def get_install_path_windows(
	program_name: str,
	ask_global: int = 0,
	add_path: bool = True,
	append_to_path: str = "",
	default_global: str = os.environ.get("ProgramFiles", "C:\\Program Files")
) -> str:
	""" Get the installation path for the program

	Args:
		program_name  (str):   The name of the program to install.
		ask_global    (int):   0 = ask for anything, 1 = install globally, 2 = install locally
		add_path      (bool):  Whether to add the program to the PATH environment variable. (Only if installed globally)
		append_to_path (str):  String to append to the installation path when adding to PATH.
			(ex: "bin" if executables are in the bin folder)
		default_global (str):  The default global installation path.
			(Default is "C:\\Program Files" which is the most common location for executables on Windows)

	Returns:
		str: The installation path.
	"""
	# Default path is located in the current working directory
	default_local_path: str = clean_path(os.path.join(os.getcwd(), program_name))

	# Define default global path (used in prompt even if not chosen initially)
	default_global_path: str = clean_path(os.path.join(default_global, program_name))

	# Ask user for installation type (global/local)
	install_type: str = ask_install_type(ask_global, default_local_path, default_global_path)

	# If the user wants to install globally,
	if install_type == 'g':

		# Check if the user has admin privileges,
		if not check_admin_windows():

			# If the user doesn't have admin privileges, fallback to local
			warning(
				f"Global installation requires administrator privileges. Please re-run as administrator.\n"
				f"Install locally instead to '{default_local_path}'? (Y/n): "
			)
			if input().lower() == 'n':
				info("Installation cancelled.")
				return ""
			else:
				# Fallback to local path if user agrees
				return prompt_for_path(
					f"Falling back to local installation path: {default_local_path}.",
					default_local_path
				)

		# If the user has admin privileges,
		else:
			# Ask it user wants to override the default global install path
			install_path: str = prompt_for_path(
				f"Default global installation path is {default_global_path}.",
				default_global_path
			)
			if add_path:
				add_to_path_windows(os.path.join(install_path, append_to_path))
			return install_path

	# Local install
	else: # install_type == 'l'
		return prompt_for_path(
			f"Default local installation path is {default_local_path}.",
			default_local_path
		)





# Imports
import time
from collections.abc import Callable
from typing import Any

from ..typing import JsonDict
from .capturer import CaptureOutput
from .common import resolve_process_title


class RemoteSubprocessError(RuntimeError):
	""" Raised in the parent when the child raised an exception - contains the child's formatted traceback. """
	def __init__(self, exc_type: str, exc_repr: str, traceback_str: str):
		msg = f"Exception in subprocess ({exc_type}): {exc_repr}\n\nRemote traceback:\n{traceback_str}"
		super().__init__(msg)
		self.remote_type = exc_type
		self.remote_repr = exc_repr
		self.remote_traceback = traceback_str


def run_in_subprocess[R](
	func: Callable[..., R],
	*args: Any,
	timeout: float | None = None,
	no_join: bool = False,
	capture_output: bool = False,
	process_title: str | None = None,
	**kwargs: Any
) -> R:
	""" Execute a function in a subprocess with positional and keyword arguments.

	This is useful when you need to run a function in isolation to avoid memory leaks,
	resource conflicts, or to ensure a clean execution environment. The subprocess will
	be created, run the function with the provided arguments, and return the result.

	Args:
		func           (Callable):     The function to execute in a subprocess.
			(SHOULD BE A TOP-LEVEL FUNCTION TO BE PICKLABLE)
		*args          (Any):          Positional arguments to pass to the function.
		timeout        (float | None): Maximum time in seconds to wait for the subprocess.
			If None, wait indefinitely. If the subprocess exceeds this time, it will be terminated.
		no_join        (bool):         If True, do not wait for the subprocess to finish (fire-and-forget) and return the Process object.
		capture_output (bool):         If True, capture the subprocess' stdout/stderr and relay it
			in real time to the parent's stdout. This enables seeing print() output
			from the subprocess in the main process.
		process_title  (str | None):   If provided, sets the process title visible in process lists.
			If it starts with '+++', this prefix is replaced by the current process title.
		**kwargs       (Any):          Keyword arguments to pass to the function.

	Returns:
		R: The return value of the function.

	Raises:
		:py:exc:`RemoteSubprocessError`: If the child raised an exception - contains the child's formatted traceback.
		:py:exc:`RuntimeError`: If the subprocess exits with a non-zero exit code or did not return a result.
		:py:exc:`TimeoutError`: If the subprocess exceeds the specified timeout.

	Examples:
		.. code-block:: python

			> # Simple function execution
			> run_in_subprocess(doctest_square, 5)
			25

			> # Function with multiple arguments
			> def add(a: int, b: int, c: int) -> int:
			.     return a + b + c
			> run_in_subprocess(add, 10, 20, c=30)
			60

			> # Function with keyword arguments
			> def greet(name: str, greeting: str = "Hello") -> str:
			.     return f"{greeting}, {name}!"
			> run_in_subprocess(greet, "World", greeting="Hi")
			'Hi, World!'

			> # With timeout to prevent hanging
			> run_in_subprocess(some_gpu_func, data, timeout=300.0, process_title="+++_gpu_worker")
	"""
	import multiprocessing as mp
	from multiprocessing import Queue

	# Create a queue to get the result from the subprocess (only if we need to wait)
	result_queue: Queue[JsonDict] | None = None if no_join else Queue()

	# Optionally setup output capture pipe and listener
	capturer: CaptureOutput | None = None
	if capture_output:
		capturer = CaptureOutput()

	# Create and start the subprocess using the module-level wrapper
	process: mp.Process = mp.Process(
		target=_subprocess_wrapper,
		args=(result_queue, func, args, kwargs),
		kwargs={"capturer": capturer, "process_title": resolve_process_title(process_title)}
	)
	process.start()

	# Function to kill the process safely
	def kill_process_tree() -> None:
		if not process.is_alive():
			return
		process.terminate()
		time.sleep(0.5)
		if process.is_alive():
			import psutil
			proc = psutil.Process(process.pid)
			procs = [proc, *proc.children(recursive=True)]
			for p in procs:
				try:
					p.terminate()
				except Exception:
					pass
			_, alive = psutil.wait_procs(procs, timeout=3)
			for p in alive:
				try:
					p.kill()
				except Exception:
					pass
		process.join()

	# For capture_output we must close the parent's copy of the write fd and start listener
	if capturer is not None:
		capturer.parent_close_write()
		capturer.start_listener()

	# Detach process if no_join (fire-and-forget)
	if result_queue is None:
		# If capturing, leave listener running in background (daemon)
		return process  # type: ignore

	# Wait for result with short polling intervals to catch KeyboardInterrupt quickly
	try:
		try:
			import queue
			start_time = time.time()
			while True:
				try:
					result_payload: JsonDict = result_queue.get(timeout=0.1)
					break
				except queue.Empty as e:
					if timeout is not None and (time.time() - start_time) >= timeout:
						raise TimeoutError(f"Subprocess exceeded timeout of {timeout} seconds and was terminated") from e
					if not process.is_alive():
						process.join()
						raise RuntimeError(f"Subprocess terminated unexpectedly with exit code {process.exitcode}") from e
		except KeyboardInterrupt:
			raise
		finally:
			# Give the child a short grace period to exit cleanly and run atexit
			# handlers (which unlink semaphores). Without this, kill_process_tree()
			# terminates the child during cleanup, leaking semaphores.
			if process.is_alive():
				process.join(timeout=2.0)
			kill_process_tree()

		# If the child sent a structured exception, raise it with the formatted traceback
		if result_payload.pop("ok", False) is False:
			raise RemoteSubprocessError(**result_payload)
		return result_payload["result"]

	# Finally, clean up queue resources and drain/join the listener
	finally:
		try:
			result_queue.cancel_join_thread()
			result_queue.close()
		except Exception:
			pass
		if capturer is not None:
			capturer.join_listener(timeout=5.0)


# "Private" function for subprocess wrapper (must be at module level for pickling on Windows)
def _subprocess_wrapper[R](
	result_queue: Any | None,
	func: Callable[..., R],
	args: tuple[Any, ...],
	kwargs: dict[str, Any],
	capturer: CaptureOutput | None = None,
	process_title: str | None = None
) -> None:
	""" Wrapper function to execute the target function and store the result in the queue.

	Must be at module level to be pickable on Windows (spawn context).

	Args:
		result_queue (multiprocessing.Queue | None):  Queue to store the result or exception (None if detached).
		func         (Callable):                      The target function to execute.
		args         (tuple):                         Positional arguments for the function.
		kwargs       (dict):                          Keyword arguments for the function.
		capturer     (CaptureOutput | None):          Optional CaptureOutput instance for stdout capture.
		process_title (str | None):                   Optional process title to set.
	"""
	try:
		# Set process title if provided
		if process_title is not None:
			import setproctitle
			setproctitle.setproctitle(process_title)

		# If a CaptureOutput instance was passed, redirect stdout/stderr to the pipe.
		if capturer is not None:
			capturer.redirect()

		# Execute the target function and put the result in the queue
		result: R = func(*args, **kwargs)
		if result_queue is not None:
			# Use timeout to prevent blocking if parent is no longer listening
			try:
				result_queue.put({"ok": True, "result": result}, timeout=5.0)
			except Exception:
				pass  # Parent likely terminated, just exit

	# Handle KeyboardInterrupt specially - don't try to send it back, just exit
	except KeyboardInterrupt:
		# Exit immediately without trying to communicate with parent
		pass

	# Handle cleanup and exceptions
	except Exception as e:
		if result_queue is not None:
			try:
				import traceback
				tb = traceback.format_exc()
				# Use timeout to prevent blocking if parent is no longer listening
				result_queue.put({
					"ok": False,
					"exc_type": e.__class__.__name__,
					"exc_repr": repr(e),
					"traceback_str": tb,
				}, timeout=5.0)
			except Exception:
				# Nothing we can do if even this fails
				pass

	finally:
		# Clean up queue to release its internal semaphores
		if result_queue is not None:
			try:
				result_queue.close()
				result_queue.join_thread()
			except Exception:
				pass

		# Restore stdout/stderr and close capturer write end
		if capturer is not None:
			capturer.child_close()  # Close child's copy of the write end




import requests
from ..decorators import handle_error as handle_error
from ..io.json import json_load as json_load
from ..io.path import clean_path as clean_path
from ..print.message import warning as warning
from collections.abc import Callable as Callable
from typing import Any

def parse_commit_message(message: str) -> tuple[str, str, str | None, bool]:
    ''' Parse a commit message following the conventional commits convention.

\tConvention format: <type>: <description> or <type>(<sub-category>): <description>

\tArgs:
\t\tmessage\t(str):\tThe commit message to parse (first line only)
\tReturns:
\t\ttuple[str, str, str | None, bool]:
\t\t\tstr:\t\tThe commit type (e.g., "Features", "Bug Fixes")
\t\t\tstr:\t\tThe commit description
\t\t\tstr | None:\tThe sub-category if present (e.g., "Project")
\t\t\tbool:\t\tTrue if it\'s a breaking change (indicated by !)
\tSource:
\t\thttps://www.conventionalcommits.org/en/v1.0.0/

\t>>> parse_commit_message("feat: Add new feature")
\t(\'Features\', \'Add new feature\', None, False)
\t>>> parse_commit_message("fix(API): Fix bug in endpoint")
\t(\'Bug Fixes\', \'Fix bug in endpoint\', \'API\', False)
\t>>> parse_commit_message("feat!: Breaking change")
\t(\'Features\', \'Breaking change\', None, True)
\t>>> parse_commit_message("docs(README): Update documentation")
\t(\'Documentation\', \'Update documentation\', \'README\', False)
\t>>> parse_commit_message("chore: Cleanup code")
\t(\'Chores\', \'Cleanup code\', None, False)
\t>>> parse_commit_message("refactor(core): Refactor module")
\t(\'Code Refactoring\', \'Refactor module\', \'core\', False)
\t>>> parse_commit_message("No conventional format")
\t(\'Other\', \'No conventional format\', None, False)
\t>>> parse_commit_message("build!: Major build change")
\t(\'Build System\', \'Major build change\', None, True)
\t>>> parse_commit_message("wip: Work in progress")
\t(\'Work in Progress\', \'Work in progress\', None, False)
\t>>> parse_commit_message("unknown_type: Some description")
\t(\'Unknowntype\', \'Some description\', None, False)
\t'''
def format_changelog(commits: list[tuple[str, str]], url_formatter: Callable[[str], str] | None = None, latest_tag_version: str | None = None, current_version: str | None = None, compare_url_formatter: Callable[[str, str], str] | None = None) -> str:
    """ Generate a changelog from a list of commits.

\tArgs:
\t\tcommits\t\t\t\t\t(list[tuple[str, str]]):\t\t\t\tList of (sha, message) tuples
\t\turl_formatter\t\t\t(Callable[[str], str] | None):\t\t\tFunction to format commit URLs.
\t\t\tTakes a SHA and returns a URL string. If None, commits show only short SHA.
\t\tlatest_tag_version\t\t(str | None):\t\t\t\t\t\t\tVersion of the previous tag for comparison link
\t\tcurrent_version\t\t\t(str | None):\t\t\t\t\t\t\tCurrent version being released
\t\tcompare_url_formatter\t(Callable[[str, str], str] | None):\t\tFunction to format comparison URL.
\t\t\tTakes (old_version, new_version) and returns a URL string.
\tReturns:
\t\tstr: Generated changelog in Markdown format
\t"""
@handle_error
def load_credentials(credentials_path: str) -> dict[str, Any]:
    ''' Load credentials from a JSON or YAML file into a dictionary.

\tLoads credentials from either a JSON or YAML file and returns them as a dictionary.
\tThe file must contain the required credentials in the appropriate format.

\tArgs:
\t\tcredentials_path (str): Path to the credentials file (.json or .yml)
\tReturns:
\t\tdict[str, Any]: Dictionary containing the credentials

\tExample JSON format:

\t.. code-block:: json

\t\t{
\t\t\t"github": {
\t\t\t\t"username": "Stoupy51",
\t\t\t\t"api_key": "ghp_XXXXXXXXXXXXXXXXXXXXXXXXXX"
\t\t\t}
\t\t}

\tExample YAML format:

\t.. code-block:: yaml

\t\tgithub:
\t\t\tusername: "Stoupy51"
\t\t\tapi_key: "ghp_XXXXXXXXXXXXXXXXXXXXXXXXXX"
\t'''
def handle_response(response: requests.Response, error_message: str) -> None:
    """ Handle a response from the API by raising an error if the response is not successful (status code not in 200-299).

\tArgs:
\t\tresponse\t\t(requests.Response): The response from the API
\t\terror_message\t(str): The error message to raise if the response is not successful
\t"""
def clean_version(version: str, keep: str = '') -> str:
    ''' Clean a version string

\tArgs:
\t\tversion\t(str): The version string to clean
\t\tkeep\t(str): The characters to keep in the version string
\tReturns:
\t\tstr: The cleaned version string

\t>>> clean_version("v1.e0.zfezf0.1.2.3zefz")
\t\'1.0.0.1.2.3\'
\t>>> clean_version("v1.e0.zfezf0.1.2.3zefz", keep="v")
\t\'v1.0.0.1.2.3\'
\t>>> clean_version("v1.2.3b", keep="ab")
\t\'1.2.3b\'
\t'''
def version_to_float(version: str, error: bool = True) -> Any:
    ''' Converts a version string into a float for comparison purposes.
\tThe version string is expected to follow the format of major.minor.patch.something_else....,
\twhere each part is separated by a dot and can be extended indefinitely.
\tSupports pre-release suffixes with numbers: devN/dN (dev), aN (alpha), bN (beta), rcN/cN (release candidate).
\tOrdering: 1.0.0 > 1.0.0rc2 > 1.0.0rc1 > 1.0.0b2 > 1.0.0b1 > 1.0.0a2 > 1.0.0a1 > 1.0.0dev1

\tArgs:
\t\tversion (str): The version string to convert. (e.g. "v1.0.0.1.2.3", "v2.0.0b2", "v1.0.0rc1")
\t\terror (bool): Return None on error instead of raising an exception
\tReturns:
\t\tfloat: The float representation of the version. (e.g. 0)

\t>>> version_to_float("v1.0.0")
\t1.0
\t>>> version_to_float("v1.0.0.1")
\t1.000000001
\t>>> version_to_float("v2.3.7")
\t2.003007
\t>>> version_to_float("v1.0.0.1.2.3")
\t1.0000000010020031
\t>>> version_to_float("v2.0") > version_to_float("v1.0.0.1")
\tTrue
\t>>> version_to_float("v2.0.0") > version_to_float("v2.0.0rc") > version_to_float("v2.0.0b") > version_to_float("v2.0.0a") > version_to_float("v2.0.0dev")
\tTrue
\t>>> version_to_float("v1.0.0b") > version_to_float("v1.0.0a")
\tTrue
\t>>> version_to_float("v1.0.0") > version_to_float("v1.0.0b")
\tTrue
\t>>> version_to_float("v3.0.0a") > version_to_float("v2.9.9")
\tTrue
\t>>> version_to_float("v1.2.3b") < version_to_float("v1.2.3")
\tTrue
\t>>> version_to_float("1.0.0") == version_to_float("v1.0.0")
\tTrue
\t>>> version_to_float("2.0.0.0.0.0.1b") > version_to_float("2.0.0.0.0.0.1a")
\tTrue
\t>>> version_to_float("2.0.0.0.0.0.1") > version_to_float("2.0.0.0.0.0.1b")
\tTrue
\t>>> version_to_float("v1.0.0rc") == version_to_float("v1.0.0c")
\tTrue
\t>>> version_to_float("v1.0.0c") > version_to_float("v1.0.0b")
\tTrue
\t>>> version_to_float("v1.0.0d") < version_to_float("v1.0.0a")
\tTrue
\t>>> version_to_float("v1.0.0dev") < version_to_float("v1.0.0a")
\tTrue
\t>>> version_to_float("v1.0.0dev") == version_to_float("v1.0.0d")
\tTrue
\t>>> version_to_float("v1.0.0rc2") > version_to_float("v1.0.0rc1")
\tTrue
\t>>> version_to_float("v1.0.0b2") > version_to_float("v1.0.0b1")
\tTrue
\t>>> version_to_float("v1.0.0a2") > version_to_float("v1.0.0a1")
\tTrue
\t>>> version_to_float("v1.0.0dev2") > version_to_float("v1.0.0dev1")
\tTrue
\t>>> version_to_float("v1.0.0") > version_to_float("v1.0.0rc2") > version_to_float("v1.0.0rc1")
\tTrue
\t>>> version_to_float("v1.0.0rc1") > version_to_float("v1.0.0b2")
\tTrue
\t>>> version_to_float("v1.0.0b1") > version_to_float("v1.0.0a2")
\tTrue
\t>>> version_to_float("v1.0.0a1") > version_to_float("v1.0.0dev2")
\tTrue
\t>>> versions = ["v1.0.0", "v1.0.0rc2", "v1.0.0rc1", "v1.0.0b2", "v1.0.0b1", "v1.0.0a2", "v1.0.0a1", "v1.0.0dev2", "v1.0.0dev1"]
\t>>> sorted_versions = sorted(versions, key=version_to_float, reverse=True)
\t>>> sorted_versions == versions
\tTrue
\t'''



"""
This module provides utility functions for upscaling images using waifu2x-ncnn-vulkan.

It includes functions to upscale individual images, batches of images in a folder,
and handle intermediate operations like converting between image formats.
The module also manages temporary directories for partial processing and tracks
progress of batch operations.

Main functionalities:

- Converting frames between image formats (PNG to JPG)
- Upscaling individual images with configurable upscale ratio
- Batch processing folders of images with progress tracking
- Handling already processed images to resume interrupted operations

Example usage:

.. code-block:: python

    from stouputils.applications.upscaler import upscale, upscale_folder

    # Upscale a single image
    upscale("input.jpg", "output.jpg", 2)

    # Upscale a folder of images
    upscale_folder("input_folder", "output_folder", 2)
"""

# Imports
import os
import shutil
import subprocess
from tempfile import TemporaryDirectory

from PIL import Image

from ...installer import check_executable
from ...io.path import clean_path
from ...parallel import multithreading
from ...print.message import debug, info
from ...print.progress_bar import colored_for_loop
from .config import WAIFU2X_NCNN_VULKAN_RELEASES, Config


# Function to convert a PNG frame to JPG format
def convert_frame(frame_path: str, delete_png: bool = True) -> None:
	""" Convert a PNG frame to JPG format to take less space.

	Args:
		frame_path  (str):   Path to the PNG frame to convert.
		delete_png  (bool):  Whether to delete the original PNG file after conversion.

	Returns:
		None: This function doesn't return anything.

	Example:
		.. code-block:: python

			> convert_frame("input.png", delete_png=True)
			> # input.png will be converted to input.jpg and the original file will be deleted

			> convert_frame("input.png", delete_png=False)
			> # input.png will be converted to input.jpg and the original file will be kept
	"""
	if frame_path.endswith(".png"):
		with Image.open(frame_path) as img:
			img.save(frame_path.replace(".png", ".jpg"), quality=Config.JPG_QUALITY)
		if delete_png:
			os.remove(frame_path)


# Function to get all frames in a folder
def get_all_files(folder: str, suffix: str | tuple[str, ...] = "") -> list[str]:
	""" Get all files paths in a folder, with a specific suffix if provided.

	Args:
		folder     (str):                    Path to the folder containing the files.
		suffix     (str | tuple[str, ...]):  Suffix of the files to get (e.g. ".png", ".jpg", etc.).

	Returns:
		list[str]: List of all files paths in the folder.

	Example:
		>>> files: list[str] = get_all_files("some_folder", ".png")
		>>> len(files)
		0
	"""
	if not os.path.exists(folder):
		return []
	return [f"{folder}/{f}" for f in os.listdir(folder) if f.endswith(suffix)]


# Function to create a temporary directory with not upscaled images
def create_temp_dir_for_not_upscaled(input_path: str, output_path: str) -> TemporaryDirectory[str] | None:
    """ Creates a temporary directory containing only images that haven't been upscaled yet.

    Args:
        input_path  (str):  Path to the folder containing input images.
        output_path (str):  Path to the folder where upscaled images are saved.

    Returns:
        TemporaryDirectory[str] | None: A temporary directory object if there are images to process,
                                        None if all images are already upscaled.
    """
    # Get all input images and the not upscaled images
    all_inputs: list[str] = get_all_files(input_path)
    not_upscaled_images: list[str] = [x for x in all_inputs if not os.path.exists(f"{output_path}/{os.path.basename(x)}")]

    # If all images or none are already upscaled, return None
    if len(not_upscaled_images) == 0 or (len(not_upscaled_images) == len(all_inputs)):
        return None

    # Create a temporary directory and copy the not upscaled images to it
    temp_dir: TemporaryDirectory[str] = TemporaryDirectory()
    debug(f"Creating temporary directory to process {len(not_upscaled_images)} images: {temp_dir.name}")
    for image in not_upscaled_images:
        shutil.copyfile(image, f"{temp_dir.name}/{os.path.basename(image)}")
    return temp_dir


# Helper function to check if the upscaler executable is installed
def check_upscaler_executable() -> None:
	if not Config.upscaler_executable_checked:
		check_executable(Config.UPSCALER_EXECUTABLE, Config.UPSCALER_EXECUTABLE_HELP_TEXT, WAIFU2X_NCNN_VULKAN_RELEASES)
		Config.upscaler_executable_checked = True


# Function to upscale an image
def upscale(input_path: str, output_path: str, upscale_ratio: int) -> None:
	""" Upscale an input image (or a directory of images) with the upscaler executable.

	Args:
		input_path     (str):  Path to the image to upscale (or a directory).
		output_path    (str):  Path to the output image (or a directory).
		upscale_ratio  (int):  Upscaling ratio.

	Example:
		.. code-block:: python

			> upscale("folder", "folder", 2)
			Traceback (most recent call last):
				...
			AssertionError: Input and output paths cannot be the same, got 'folder'

			> upscale("stouputils", "stouputils/output.jpg", 2)
			Traceback (most recent call last):
				...
			AssertionError: If input is a directory, output must be a directory too, got 'stouputils/output.jpg'


			> upscale("input.jpg", "output.jpg", 2)
			> # The input.jpg will be upscaled to output.jpg with a ratio of 2

			> upscale("input_folder", "output_folder", 2)
			> # The input_folder will be upscaled to output_folder with a ratio of 2
	"""
	check_upscaler_executable()
	is_input_dir: bool = os.path.isdir(input_path)
	is_output_dir: bool = os.path.isdir(output_path)

	# Assertions
	assert input_path != output_path, f"Input and output paths cannot be the same, got '{input_path}'"
	invalid_dir_combination: bool = is_input_dir == True and is_output_dir == False  # noqa: E712
	assert not invalid_dir_combination, f"If input is a directory, output must be a directory too, got '{output_path}'"

	# Convert output_path to a file path if it's a directory
	if is_output_dir and not is_input_dir:

		# Needs to be a PNG file to be converted to JPG later
		output_file_name: str = os.path.basename(input_path).replace(".jpg", ".png")
		output_path = clean_path(f"{output_path}/{output_file_name}")
		is_output_dir = False

	# If both input and output are folders, and there are images already upscaled in the output folder,
	# Then create a temporary folder with not upscaled images
	temp_dir: TemporaryDirectory[str] | None = None
	if is_input_dir and is_output_dir:
		temp_dir = create_temp_dir_for_not_upscaled(input_path, output_path)
		if temp_dir:
			input_path = temp_dir.name

	# Build the command and run it
	cmd: list[str] = [Config.UPSCALER_EXECUTABLE, *Config.UPSCALER_ARGS]
	cmd[cmd.index("INPUT_PATH")] = input_path	          # Replace the input path
	cmd[cmd.index("OUTPUT_PATH")] = output_path	          # Replace the output path
	cmd[cmd.index("UPSCALE_RATIO")] = str(upscale_ratio)  # Replace the upscaled ratio (if using waifu2x-ncnn-vulkan)
	subprocess.run(cmd, capture_output=True)

	# If the input was a temporary folder, delete it
	if temp_dir:
		temp_dir.cleanup()

	# Convert the output frames to JPG format
	if not is_output_dir:
		convert_frame(output_path)
	else:
		frames_to_convert: list[str] = get_all_files(output_path, ".png")
		if frames_to_convert:
			multithreading(convert_frame, frames_to_convert, desc="Converting frames to JPG format")


# Function to upscale multiple images
def upscale_images(images: list[str], output_folder: str, upscale_ratio: int, desc: str = "Upscaling images") -> None:
	""" Upscale multiple images from a list.

	Args:
		images        (list[str]): List of paths to the images to upscale.
		output_folder (str):       Path to the output folder where the upscaled images will be saved.
		upscale_ratio (int):       Upscaling ratio.
		desc          (str):       Description of the function execution displayed in the progress bar.
			No progress bar will be displayed if desc is empty.

	Returns:
		None: This function doesn't return anything.
	"""
	for image_path in (colored_for_loop(images, desc=desc) if desc != "" else images):
		upscale(image_path, output_folder, upscale_ratio)

# Function to upscale a folder of images
def upscale_folder(
	input_folder: str,
	output_folder: str,
	upscale_ratio: int,
	slightly_faster_mode: bool = True,
	desc: str = "Upscaling folder"
) -> None:
	""" Upscale all images in a folder.

	Args:
		input_folder          (str):   Path to the input folder containing the images to upscale.
		output_folder         (str):   Path to the output folder where the upscaled images will be saved.
		upscale_ratio         (int):   Upscaling ratio.
		slightly_faster_mode  (bool):  Whether to use the slightly faster mode (no progress bar),
			one call to the upscaler executable.
		desc                  (str):   Description of the function execution displayed in the progress bar.
			No progress bar will be displayed if desc is empty.

	Returns:
		None: This function doesn't return anything.
	"""
	info(f"Upscaling '{input_folder}' to '{output_folder}' with a ratio of {upscale_ratio}...")
	if slightly_faster_mode:
		upscale(input_folder, output_folder, upscale_ratio)
	else:
		inputs: list[str] = get_all_files(input_folder)
		not_upscaled_images: list[str] = [x for x in inputs if not os.path.exists(f"{output_folder}/{os.path.basename(x)}")]
		upscale_images(not_upscaled_images, output_folder, upscale_ratio, desc=desc)




"""
This module provides utilities for typing enhancements such as JSON type aliases:

- :py:class:`JsonDict`
- :py:class:`JsonList`
- :py:class:`JsonMap`
- :py:class:`JsonMutMap`
- :py:class:`IterAny`
- :py:func:`convert_to_serializable`
"""

# Imports
from collections.abc import Callable, Iterable, Mapping, MutableMapping
from dataclasses import asdict, is_dataclass
from typing import Any, cast

# Typing aliases
type JsonDict = dict[str, Any]
""" A type alias for JSON dictionaries """
type JsonList = list[Any]
""" A type alias for JSON lists """
type JsonMap = Mapping[str, Any]
""" A type alias for JSON mapping """
type JsonMutMap = MutableMapping[str, Any]
""" A type alias for mutable JSON mapping """
type IterAny = Iterable[Any]
""" A type alias for iterable of any type """
type CallableAny = Callable[..., Any]
""" A type alias for any callable """


## Utility functions
def convert_to_serializable(obj: Any) -> Any:
	""" Recursively convert objects to JSON-serializable forms.

	Objects with a `to_dict()` or `asdict()` method are converted to their dictionary representation.
	Dictionaries and lists are recursively processed.

	Can also be used to convert nested structures containing custom objects,
	such as defaultdict, dataclasses, or other user-defined types.

	Args:
		obj (Any): The object to convert
	Returns:
		Any: The JSON-serializable version of the object
	Examples:
		>>> from typing import defaultdict
		>>> my_dict = defaultdict(lambda: defaultdict(int))
		>>> my_dict['a']['b'] += 6
		>>> my_dict['c']['d'] = 4
		>>> my_dict['a']
		defaultdict(<class 'int'>, {'b': 6})
		>>> my_dict['c']
		defaultdict(<class 'int'>, {'d': 4})
		>>> convert_to_serializable(my_dict)
		{'a': {'b': 6}, 'c': {'d': 4}}

		>>> from dataclasses import dataclass
		>>> @dataclass
		... class Point:
		...     x: int
		...     y: int
		>>> convert_to_serializable(Point(3, 4))
		{'x': 3, 'y': 4}
	"""
	if hasattr(obj, "to_dict"):
		return obj.to_dict()
	elif is_dataclass(obj):
		return asdict(obj) # pyright: ignore[reportArgumentType]
	elif isinstance(obj, dict | Mapping | MutableMapping):
		return {k: convert_to_serializable(v) for k, v in cast(JsonDict, obj).items()}
	elif isinstance(obj, Iterable) and not isinstance(obj, (str, bytes)):
		return [convert_to_serializable(item) for item in cast(IterAny, obj)]
	return obj




from ..print.message import warning as warning
from .common import get_function_name as get_function_name, get_wrapper_name as get_wrapper_name, set_wrapper_name as set_wrapper_name
from collections.abc import Callable as Callable
from typing import Any, overload

@overload
def retry[T](func: Callable[..., T], *, exceptions: tuple[type[BaseException], ...] | type[BaseException] = ..., max_attempts: int | None = 10, delay: float = 1.0, backoff: float = 1.0, message: str = '', on_each_failure: Callable[[BaseException, int], Any] | None = None) -> Callable[..., T]: ...
@overload
def retry[T](func: None = None, *, exceptions: tuple[type[BaseException], ...] | type[BaseException] = ..., max_attempts: int | None = 10, delay: float = 1.0, backoff: float = 1.0, message: str = '', on_each_failure: Callable[[BaseException, int], Any] | None = None) -> Callable[[Callable[..., T]], Callable[..., T]]: ...



"""
This module provides functions for creating and managing archives.

- :py:func:`~repair_zip_file.repair_zip_file` - Try to repair a corrupted zip file by ignoring some of the errors
- :py:func:`~make_archive.make_archive` - Create a zip archive from a source directory with consistent file timestamps.
- :py:func:`~cli.archive_cli` - Main entry point for command line usage

.. image:: https://raw.githubusercontent.com/Stoupy51/stouputils/refs/heads/main/assets/archive_module.gif
  :alt: stouputils archive examples
"""

# Imports
from .cli import *
from .make_archive import *
from .repair_zip_file import *  # pyright: ignore[reportGeneralTypeIssues]

if __name__ == "__main__":
	archive_cli()




from .path import clean_path as clean_path

def is_junction(path: str) -> bool:
    """ Check if a path is a junction point (Windows) or a symlink (any OS).

\tArgs:
\t\tpath (str): The path to check
\tReturns:
\t\tbool: True if the path is a junction or symlink
\t"""
def create_junction(source: str, target: str) -> None:
    """ Create a directory junction on Windows pointing source -> target.
\tUses 'mklink /J' command.

\tArgs:
\t\tsource (str): The junction path to create (with forward slashes, will be converted)
\t\ttarget (str): The target directory the junction points to
\t"""
def create_bind_mount(source: str, target: str) -> None:
    """ Create a bind mount on Linux pointing source -> target.
\tUses ``mount --bind`` command (requires root/sudo).

\tNote:
\t\tBind mounts do **not** persist across reboots unless an entry is added to ``/etc/fstab``.
\t\tTo make it persistent, add this line to ``/etc/fstab``::

\t\t\t/absolute/path/to/target /absolute/path/to/source none bind 0 0

\tArgs:
\t\tsource (str): The mount point path to create (must exist as an empty directory)
\t\ttarget (str): The target directory to bind
\tRaises:
\t\t:py:exc:`OSError`: If the bind mount command fails
\t"""
def copytree_with_progress(source: str, destination: str, desc: str = 'Copying') -> str:
    ''' Copy a directory tree from source to destination with a colored progress bar.

\tUses :func:`~stouputils.print.progress_bar.colored_for_loop` to display progress while copying each file.
\tDirectory structure is created automatically. Existing files at the destination are overwritten.

\tArgs:
\t\tsource\t\t(str):\tPath to the source directory to copy
\t\tdestination\t(str):\tPath to the destination directory
\t\tdesc\t\t(str):\tDescription for the progress bar (default: ``"Copying"``)
\tReturns:
\t\tstr:\tThe destination path
\tRaises:
\t\t:py:exc:`NotADirectoryError`: If source is not a directory

\tExamples:

\t\t.. code-block:: python

\t\t\t> copytree_with_progress("C:/Games/MyGame", "D:/Backup/MyGame")
\t\t\t# Copying: 100%|██████████████████| 150/150 [00:05<00:00, 30.00it/s]
\t\t\t\'D:/Backup/MyGame\'
\t'''
def redirect_folder(source: str, destination: str, link_type: str | None = None) -> str:
    ''' Move a folder from source to destination and create a link at the original source location.

\tIf the source is already a symlink or junction, the operation is skipped.
\tIf the destination path ends with ``/``, the source folder\'s basename is appended automatically.

\tLink types:
\t\t- ``"junction"`` (or ``"hardlink"``): Uses an NTFS junction on Windows,
\t\t\tor a bind mount (``mount --bind``, requires sudo) on Linux.
\t\t\tFalls back to symlink if junction/bind mount creation fails.
\t\t- ``"symlink"``: Uses a symbolic link (may require elevated privileges on Windows).
\t\t- ``None``: Prompts the user interactively.

\tArgs:
\t\tsource\t\t(str):\t\t\t\tPath to the existing folder to redirect
\t\tdestination\t(str):\t\t\t\tPath where the folder contents will be moved to
\t\tlink_type\t(str | None):\t\t``"hardlink"``/``"junction"``, ``"symlink"``, or None to ask
\tReturns:
\t\tstr:\tThe final destination path

\tExamples:

\t\t.. code-block:: python

\t\t\t> redirect_folder("C:/Games/MyGame", "D:/Games/")
\t\t\t# Moves C:/Games/MyGame -> D:/Games/MyGame and creates a link at C:/Games/MyGame

\t\t\t> redirect_folder("C:/Games/MyGame", "D:/Storage/MyGame")
\t\t\t# Moves C:/Games/MyGame -> D:/Storage/MyGame and creates a link at C:/Games/MyGame
\t'''
def redirect_cli() -> None:
    """ CLI entry point for the redirect command.
\tUsage: stouputils redirect <source> <destination> [--hardlink|--symlink]
\t"""




# Imports
from collections.abc import Callable
from functools import wraps
from traceback import format_exc
from typing import Any, overload

from ..print.message import error, warning
from .common import get_function_name, get_wrapper_name, set_wrapper_name
from .handle_error import LogLevels


# Decorator that marks a function as deprecated
@overload
def deprecated[T](
	func: Callable[..., T],
	*,
	message: str = "",
	version: str = "",
	error_log: LogLevels = LogLevels.WARNING
) -> Callable[..., T]: ...

@overload
def deprecated[T](
	func: None = None,
	*,
	message: str = "",
	version: str = "",
	error_log: LogLevels = LogLevels.WARNING
) -> Callable[[Callable[..., T]], Callable[..., T]]: ...

def deprecated[T](
	func: Callable[..., T] | None = None,
	*,
	message: str = "",
	version: str = "",
	error_log: LogLevels = LogLevels.WARNING
) -> Callable[..., T] | Callable[[Callable[..., T]], Callable[..., T]]:
	""" Decorator that marks a function as deprecated.

	Args:
		func        (Callable[..., T] | None): Function to mark as deprecated
		message     (str):                       Additional message to display with the deprecation warning
		version     (str):                       Version since when the function is deprecated (e.g. "v1.2.0")
		error_log   (LogLevels):                 Log level for the deprecation warning

			- :attr:`LogLevels.NONE` - None
			- :attr:`LogLevels.WARNING` - Show as warning
			- :attr:`LogLevels.WARNING_TRACEBACK` - Show as warning with traceback
			- :attr:`LogLevels.ERROR_TRACEBACK` - Show as error with traceback
			- :attr:`LogLevels.RAISE_EXCEPTION` - Raise exception

	Examples:
		>>> @deprecated
		... def old_function():
		...     pass

		>>> @deprecated(message="Use 'new_function()' instead", error_log=LogLevels.WARNING)
		... def another_old_function():
		...     pass
	"""
	def decorator(func: Callable[..., T]) -> Callable[..., T]:
		@wraps(func)
		def wrapper(*args: tuple[Any, ...], **kwargs: dict[str, Any]) -> Any:
			# Build deprecation message
			msg: str = f"Function '{get_function_name(func)}()' is deprecated"
			if version:
				msg += f" since {version}"
			if message:
				msg += f". {message}"

			# Handle deprecation warning based on log level
			if error_log == LogLevels.WARNING:
				warning(msg)
			elif error_log == LogLevels.WARNING_TRACEBACK:
				warning(f"{msg}\n{format_exc()}")
			elif error_log == LogLevels.ERROR_TRACEBACK:
				error(f"{msg}\n{format_exc()}", exit=True)
			elif error_log == LogLevels.RAISE_EXCEPTION:
				raise DeprecationWarning(msg)

			# Call the original function
			return func(*args, **kwargs)
		set_wrapper_name(wrapper, get_wrapper_name("stouputils.decorators.deprecated", func))
		return wrapper

	# Handle both @deprecated and @deprecated(message=..., error_log=...)
	if func is None:
		return decorator
	return decorator(func)




"""
This module provides utility functions for printing messages with different levels of importance.

If a message is printed multiple times, it will be displayed as "(xN) message"
where N is the number of times the message has been printed.

The module also includes a :py:func:`~color_formatting.colored` function that formats text with Python 3.14 style coloring
for file paths, line numbers, function names (in magenta), and exception names (in bold magenta).
All functions have their colored counterparts with a 'c' suffix (e.g., :py:func:`infoc`, :py:func:`debugc`, etc.)

.. image:: https://raw.githubusercontent.com/Stoupy51/stouputils/refs/heads/main/assets/print_module.gif
  :alt: stouputils print examples
"""

# Imports
from .color_formatting import *
from .common import *
from .debugging import *
from .message import *
from .output_stream import *
from .progress_bar import *
from .utils import *

# Test the print functions
if __name__ == "__main__":
	info("Hello", "World")
	time.sleep(0.5)
	info("Hello", "World")
	time.sleep(0.5)
	info("Hello", "World")
	time.sleep(0.5)
	info("Not Hello World !")
	time.sleep(0.5)
	info("Hello", "World")
	time.sleep(0.5)
	info("Hello", "World")

	# All remaining print functions
	alt_debug("Hello", "World")
	debug("Hello", "World")
	suggestion("Hello", "World")
	progress("Hello", "World")
	warning("Hello", "World")
	error("Hello", "World", exit=False)
	whatisit("Hello")
	whatisit("Hello", "World")

	# Test whatisit with different types
	import numpy as np
	print()
	whatisitc(
		123,
		"Hello World",
		[1, 2, 3, 4, 5],
		np.array([[1, 2, 3], [4, 5, 6]]),
		{"a": 1, "b": 2},
	)




from ..decorators.measure_time import measure_time as measure_time
from ..io.path import clean_path as clean_path, relative_path as relative_path
from ..print.message import error as error, info as info, warning as warning
from .utils import test_module_with_progress as test_module_with_progress
from doctest import TestResults as TestResults

def launch_tests(root_dir: str, strict: bool = True, pattern: str = '*') -> int:
    ''' Main function to launch tests for all modules in the given directory.

\tArgs:
\t\troot_dir\t(str):\t\tRoot directory to search for modules
\t\tstrict\t\t(bool):\t\tModify the FORCE_RAISE_EXCEPTION configuration to True
\t\tpattern\t\t(str):\t\tPattern to filter module names (fnmatch style, e.g., \'*typ*\', \'io\', etc.)

\tReturns:
\t\tint: The number of failed tests

\tExamples:
\t\t>>> launch_tests("unknown_dir")
\t\tTraceback (most recent call last):
\t\t\t...
\t\tValueError: No modules found in \'unknown_dir\'

\t.. code-block:: python

\t\t> if launch_tests("/path/to/source") > 0:
\t\t\tsys.exit(1)
\t\t[PROGRESS HH:MM:SS] Importing module \'module1\'\ttook 0.001s
\t\t[PROGRESS HH:MM:SS] Importing module \'module2\'\ttook 0.002s
\t\t[PROGRESS HH:MM:SS] Importing module \'module3\'\ttook 0.003s
\t\t[PROGRESS HH:MM:SS] Importing module \'module4\'\ttook 0.004s
\t\t[INFO HH:MM:SS] Testing 4 modules...
\t\t[PROGRESS HH:MM:SS] Testing module \'module1\'\ttook 0.005s
\t\t[PROGRESS HH:MM:SS] Testing module \'module2\'\ttook 0.006s
\t\t[PROGRESS HH:MM:SS] Testing module \'module3\'\ttook 0.007s
\t\t[PROGRESS HH:MM:SS] Testing module \'module4\'\ttook 0.008s
\t'''



from ..typing import JsonDict as JsonDict
from .capturer import CaptureOutput as CaptureOutput
from .common import resolve_process_title as resolve_process_title
from _typeshed import Incomplete
from collections.abc import Callable as Callable
from typing import Any

class RemoteSubprocessError(RuntimeError):
    """ Raised in the parent when the child raised an exception - contains the child's formatted traceback. """
    remote_type: Incomplete
    remote_repr: Incomplete
    remote_traceback: Incomplete
    def __init__(self, exc_type: str, exc_repr: str, traceback_str: str) -> None: ...

def run_in_subprocess[R](func: Callable[..., R], *args: Any, timeout: float | None = None, no_join: bool = False, capture_output: bool = False, process_title: str | None = None, **kwargs: Any) -> R:
    ''' Execute a function in a subprocess with positional and keyword arguments.

\tThis is useful when you need to run a function in isolation to avoid memory leaks,
\tresource conflicts, or to ensure a clean execution environment. The subprocess will
\tbe created, run the function with the provided arguments, and return the result.

\tArgs:
\t\tfunc           (Callable):     The function to execute in a subprocess.
\t\t\t(SHOULD BE A TOP-LEVEL FUNCTION TO BE PICKLABLE)
\t\t*args          (Any):          Positional arguments to pass to the function.
\t\ttimeout        (float | None): Maximum time in seconds to wait for the subprocess.
\t\t\tIf None, wait indefinitely. If the subprocess exceeds this time, it will be terminated.
\t\tno_join        (bool):         If True, do not wait for the subprocess to finish (fire-and-forget) and return the Process object.
\t\tcapture_output (bool):         If True, capture the subprocess\' stdout/stderr and relay it
\t\t\tin real time to the parent\'s stdout. This enables seeing print() output
\t\t\tfrom the subprocess in the main process.
\t\tprocess_title  (str | None):   If provided, sets the process title visible in process lists.
\t\t\tIf it starts with \'+++\', this prefix is replaced by the current process title.
\t\t**kwargs       (Any):          Keyword arguments to pass to the function.

\tReturns:
\t\tR: The return value of the function.

\tRaises:
\t\t:py:exc:`RemoteSubprocessError`: If the child raised an exception - contains the child\'s formatted traceback.
\t\t:py:exc:`RuntimeError`: If the subprocess exits with a non-zero exit code or did not return a result.
\t\t:py:exc:`TimeoutError`: If the subprocess exceeds the specified timeout.

\tExamples:
\t\t.. code-block:: python

\t\t\t> # Simple function execution
\t\t\t> run_in_subprocess(doctest_square, 5)
\t\t\t25

\t\t\t> # Function with multiple arguments
\t\t\t> def add(a: int, b: int, c: int) -> int:
\t\t\t.     return a + b + c
\t\t\t> run_in_subprocess(add, 10, 20, c=30)
\t\t\t60

\t\t\t> # Function with keyword arguments
\t\t\t> def greet(name: str, greeting: str = "Hello") -> str:
\t\t\t.     return f"{greeting}, {name}!"
\t\t\t> run_in_subprocess(greet, "World", greeting="Hi")
\t\t\t\'Hi, World!\'

\t\t\t> # With timeout to prevent hanging
\t\t\t> run_in_subprocess(some_gpu_func, data, timeout=300.0, process_title="+++_gpu_worker")
\t'''
def _subprocess_wrapper[R](result_queue: Any | None, func: Callable[..., R], args: tuple[Any, ...], kwargs: dict[str, Any], capturer: CaptureOutput | None = None, process_title: str | None = None) -> None:
    """ Wrapper function to execute the target function and store the result in the queue.

\tMust be at module level to be pickable on Windows (spawn context).

\tArgs:
\t\tresult_queue (multiprocessing.Queue | None):  Queue to store the result or exception (None if detached).
\t\tfunc         (Callable):                      The target function to execute.
\t\targs         (tuple):                         Positional arguments for the function.
\t\tkwargs       (dict):                          Keyword arguments for the function.
\t\tcapturer     (CaptureOutput | None):          Optional CaptureOutput instance for stdout capture.
\t\tprocess_title (str | None):                   Optional process title to set.
\t"""
